#include "../include/agt_alloc.h"

/* ********************************************************/
/* Memory allocation and freeing **************************/
/* ********************************************************/

struct node* allocClusters(int nb_clusters, int x_len, int y_len)
{
  struct node *clust; 
  int k;
  
  /* allocate clusters */
  clust = (struct node*)malloc(sizeof(struct node)*nb_clusters);

  if (clust==NULL)
    {
      fprintf(stderr,"Can't allocate clusters\a\n");
      exit(-1);
    }


  for (k=0;k<nb_clusters;k++)
    {
      clust[k].root_FLAG = 1;
      clust[k].prune_FLAG = -1; /* means no pruning, off */
      clust[k].parent = clust[k].child1 = clust[k].child2 = -1;
      clust[k].nb_points=0;
      clust[k].mean = (double*)malloc(sizeof(double)*x_len);
      if ((clust[k].mean==NULL)) 
	{
	  fprintf(stderr,"Can't allocate cluster means or Cxe_sig\a\n");
	  exit(-1);
	}

      /* alloc cluster stats also */
      allocClusterStats(&clust[k], x_len,y_len);
      
      clust[k].a = (double**)multialloc(sizeof(double),2,x_len,y_len);
      clust[k].b = (double*)malloc(sizeof(double)*y_len);
      if (clust[k].b==NULL)
	{
	  fprintf(stderr,"Malloc error\n\a\n");
	  exit(-1);
	}

    }
  
  return(clust);
}



void freeClusters(struct node *clust, int nb_clusters)
{
  int k;

  for (k=0;k<nb_clusters;k++)
    {
      free((void*)clust[k].mean);
      freeClusterStatsandFlt(&clust[k]);
   }
  free((void*)clust);
}




void allocClusterStats(struct node *clust, int x_len, int y_len)
{
  clust->R = (double**)multialloc(sizeof(double),2,x_len,x_len);
  clust->Cyy = (double**)multialloc(sizeof(double),2,y_len,y_len);
  clust->vect = (double**)multialloc(sizeof(double),2,x_len,y_len);
  clust->xmean = (double*)malloc(sizeof(double)*x_len);
        

  clust->yer2 = (double**)multialloc(sizeof(double),2,y_len,y_len);


  if (clust->xmean==NULL)
    {
      fprintf(stderr,"Malloc error\n\a\n");
      exit(-1);
    }
  

  clust->ymean = (double*)malloc(sizeof(double)*y_len);
  if (clust->ymean==NULL)
    {
      fprintf(stderr,"Malloc error\n\a\n");
      exit(-1);
    }

  clear_clust_stats(clust,x_len,y_len);

}



void clear_clust_stats(struct node *clust, int x_len, int y_len)
{
  int i,j;
  
  for (i=0;i<x_len;i++)
    for (j=0;j<x_len;j++)
      clust->R[i][j]=(double)0.0;
  for (i=0;i<x_len;i++)
    for (j=0;j<y_len;j++)
      clust->vect[i][j]=(double)0.0;

  for (i=0;i<y_len;i++)
    for (j=0;j<y_len;j++)
      clust->Cyy[i][j]=(double)0.0;

  for (j=0;j<x_len;j++)
    clust->xmean[j]=(double)0.0;

  for (j=0;j<y_len;j++)
    for (i=0;i<y_len;i++)
      clust->yer2[j][i]=(double)0.0;

  for (j=0;j<y_len;j++)
    clust->ymean[j]=(double)0.0;
  
}
 





void freeClusterStatsandFlt(struct node *clust)
{
  multifree((void**)clust->R,2);
  multifree((void**)clust->Cyy,2);
  multifree((void**)clust->vect,2);
  multifree((void**)clust->a,2);
  free((void*)clust->xmean);
  multifree((void*)clust->yer2,2);
  free((void*)clust->ymean);
  free((void*)clust->b);
}


