#include "../include/agt_clust.h"

int determine_cluster(double *xvec, int x_len, struct node *clust, struct cmds *params)
{ 
  double d,mind;
  int best_cluster;
  int k;
  int nb_clusters;
  
  nb_clusters = params->nb_clusters;

  d=dist(xvec, clust[0].mean, x_len);
  mind=d;
  best_cluster=0;
	  
  for (k=1;k<nb_clusters;k++)
    {
      d=dist(xvec, clust[k].mean, x_len);
      if (d<mind)
	{
	  mind=d;
	  best_cluster=k;
	}
    }
  

  return(best_cluster);

}
  
void classifyData(struct node *clust, struct data *train, struct cmds *params, int nb_iter)
{
  int nb_points,x_len,nb_clusters,i,j,k,c;
  
  double cumd=0.0,old_cumd=0.0;
  int best_cluster;
  int iter;

  /*int *points;*/
  double **new_mean;

  nb_points=train->nb_points;
  x_len = train->x_len;
  nb_clusters = params->nb_clusters;
  

 
  new_mean = (double**)multialloc(sizeof(double),2,nb_clusters,x_len);
  
  

  /* Main VQ loop */
  for (iter=0;iter<nb_iter;iter++)
    {

      old_cumd=cumd;
      cumd=0.0;


      /* clear new cluster statistics */
      for (k=0;k<nb_clusters;k++)
	{
	  /*points[k]=0;*/
	  clust[k].nb_points=0;
	  for (j=0;j<x_len;j++)
	    new_mean[k][j] = 0.0;
	}
      
      
      /* Assign all data points to clusters */
      for (i=0;i<nb_points;i++)
	{

	  best_cluster=determine_cluster(train->data[i], x_len, clust, params);
	  
	  cumd = cumd + dist(train->data[i], clust[best_cluster].mean, x_len);
	  
	  
	  /* add data point to cluster mean */
	  for (j=0;j<x_len;j++)
	    new_mean[best_cluster][j] += train->data[i][j]; 
	  clust[best_cluster].nb_points = clust[best_cluster].nb_points +1;
	  
	}


      /* Write normalized mean and #points into cluster struct */
      for (k=0;k<nb_clusters;k++)
	{
	  if (clust[k].nb_points > 0)
	    for (j=0;j<x_len;j++)
	      clust[k].mean[j] = new_mean[k][j]/((double)clust[k].nb_points);
	}


     
      /* eliminate empty clusters */
      k=0;
      for (;;)
	{
	  if (clust[k].nb_points==0)
	    {
	      nb_clusters--;
	      params->nb_clusters = nb_clusters;
	      printf("Cluster %d empty, eliminated, nb_clusters=%d\n",k,params->nb_clusters);
	      for (c=k;c<nb_clusters;c++) 
		{
		  clust[c].root_FLAG = clust[c+1].root_FLAG;
		  clust[c].parent = clust[c+1].parent;
		  clust[c].child1 = clust[c+1].child1;
		  clust[c].child2 = clust[c+1].child2;
		  clust[c].nb_points = clust[c+1].nb_points;
		  for (j=0;j<x_len;j++)
		    clust[c].mean[j] =  clust[c+1].mean[j];
		}
	    }
	  else
	    k++;
	  
	  if (k>=nb_clusters)
	    break;
	} 
      


      printf("cum_dist = %f\n",cumd);

      if ((cumd-old_cumd>1e-8) && (iter>0))
	{
	  fprintf(stderr,"Error: cumd =%g > old_cumd =%g\a\n\n",cumd,old_cumd);
	  exit(-1);
	}



      if ((iter>0) && (fabs(cumd-old_cumd)<1e-30))
	break;
    } /* end for iter */
      
  multifree((void**)new_mean,2);
  
      
}

void cluster_data(int *clustered_data, struct node *clust, struct data *train, struct cmds *params)   
{ 
  /* routine is used when computing node statistics for filters */
  int nb_points,x_len;
  int best_cluster,i;
  
  nb_points=train->nb_points;

  x_len = train->x_len;

  for (i=0;i<nb_points;i++)
    {
      best_cluster=determine_cluster(train->data[i], x_len, clust, params);
      clustered_data[i] = best_cluster;
    }
}


double dist(double *vec1, double *vec2, int len)
{
  double d;
  double x;
  int i;

  d=(double)0.0;

  for (i=0;i<len;i++)
    {
      x = vec1[i]-vec2[i];
      d += (x*x);
    }

  return(d);
}



void initClmeans(struct node *clust, struct data *train, struct cmds *params)
{
  int nb_points,x_len,nb_clusters,i,j,k;
  int same=0,new,r;
  double dnb_points;
  double *vec;


  nb_points=train->nb_points;
  dnb_points = (double)nb_points;
  x_len = train->x_len;
  nb_clusters = params->nb_clusters;

  
  vec=(double*)malloc(sizeof(double)*x_len);
  if (vec==NULL)
    {
      fprintf(stderr,"malloc error\a\n");
      exit(-1);
    }

  for (k=0;k<nb_clusters;k++)
    {
      new=0;
      while (new==0)
	{
	  /* randomly select data point */
	  r=(int)(random2()*dnb_points);

	  /* copy data point in vec */
	  for (j=0;j<x_len;j++)
	    vec[j]=train->data[r][j];

	  /* check whether previous cluster has same mean */
	  for (i=0;i<k;i++)
	    {
	      /* compare vec to mean of cluster i */
	      same = 1;
	      for (j=0;j<x_len;j++)
		if (fabs(clust[i].mean[j] - vec[j]) > 1e-7) 
		  {
		    same=0;
		    break; /* if one coordinate different, break */
		  }
	     
	      /* if no coordinate different -> same=1, */
	      /* i.e. need to try again */
	      if (same==1)
		break;
	    }
	  if (same==0)
	    new=1;
	
	} /* end while new==0 */
      
      /* Now that we have new point, copy in clust[k].mean */
      /*printf("initcluster %d  to point %d ",k,r);*/
      for (j=0;j<x_len;j++)
	{
	  clust[k].mean[j] = vec[j];
	}

    } /* end for k */

 

  free((void*)vec);


}




void test_trset_size(struct data *train, struct cmds *params)
{
  int x_len,nb_clusters,pt,k,j;
  int same=0;
  int no_distinct_pts=1; 

  x_len = train->x_len;
  nb_clusters = params->nb_clusters;

  for (pt=1;pt<train->nb_points;pt++)
    {
      for (k=0;k<pt;k++)
	{
	  same=1;
	  for (j=0;j<x_len;j++)
	    if (fabs(train->data[pt][j] - train->data[k][j]) > 1e-7) 
	      {
		same=0;
		break; /* if one coordinate different, break */
	      }
	  if (same==1)
	    break;
	}
      if (same==0)
	no_distinct_pts++;
      
      if (no_distinct_pts>(int)(1.2*nb_clusters+1))
	break;
	
    }

  if (no_distinct_pts<(int)(1.2*nb_clusters))
    {
      fprintf(stderr,"\n\nError: The train data set does not have enough\ndistinct data points for specified number\nof initial clusters (leaves).\nReduce the number of initial clusters to below\napproximately %d or increase size of the training set.\n\a\n",(int)(no_distinct_pts/1.2));
      exit(-1);
    }
  




}
