#include "../include/agt_filter.h"

/* ********************************************************/
/* ****** Routines to compute node predictors  ************/
/* ********************************************************/

void compute_filters(struct node *clust, struct data *train,  struct cmds *params)
{
  int nb_clusters; 
  int k;

  nb_clusters = params->nb_clusters;
  compute_clust_stats(clust, train, params);
  
  for (k=0;k<nb_clusters;k++)
    {
      if (clust[k].nb_points>0)
	{
	  compute_clust_filter(&clust[k],train);
	  compute_clust_prederr(&clust[k],train);
	}
      else
	{
	  clear_clust_stats(&clust[k], train->x_len, train->y_len);
	  clust[k].pred_err=0.0;
	}
    } 

}
void compute_clust_stats(struct node *clust, struct data *train, struct cmds *params)
{
  int nb_points,nb_clusters,x_len, y_len;
  int i,j,l,cl;
  int *clustered_data; /* cluster id for each data point */


  double nbpts_dbl;
  
  nb_clusters = params->nb_clusters;
  nb_points=train->nb_points;
  x_len = train->x_len;
  y_len = train->y_len;
  
  clustered_data = (int*)malloc(sizeof(int)*nb_points);
  if (clustered_data == NULL)
    {
      fprintf(stderr,"malloc error\a\n");
      exit(-1);
    }
  

  /* assign each data point to a cluster */
  cluster_data(clustered_data, clust, train, params);

  

  /* clear all cluster statistics */
  for (cl=0;cl<nb_clusters;cl++)
    {
      clear_clust_stats(&clust[cl],x_len,y_len);
      clust[cl].nb_points=0;
    }   
 
  /* clear x-mean */
  for (cl=0;cl<nb_clusters;cl++)
    { 
      for (i=0;i<x_len;i++)
	clust[cl].xmean[i] = 0.0;
    }

  /* compute nb_points */
  for (l=0;l<nb_points;l++)
    clust[clustered_data[l]].nb_points = clust[clustered_data[l]].nb_points + 1;

  /* comp x-mean */
  for (l=0;l<nb_points;l++)
    for (i=0;i<x_len;i++)
      clust[clustered_data[l]].xmean[i] += (train->data[l][i]);

  for (cl=0;cl<nb_clusters;cl++)
    if (clust[cl].nb_points>0)
      for (i=0;i<x_len;i++)
	clust[cl].xmean[i] /= (double)clust[cl].nb_points;



  /* clear old ymean and mean */
  for (cl=0;cl<nb_clusters;cl++)
    { 
      for (i=0;i<y_len;i++)
       clust[cl].ymean[i] = 0.0;
    }

  /* compute y-mean */
  for (l=0;l<nb_points;l++)
    for (i=0;i<y_len;i++)
      clust[clustered_data[l]].ymean[i] += (train->data[l][i+x_len]);

  for (cl=0;cl<nb_clusters;cl++)
    if (clust[cl].nb_points>0)
      for (i=0;i<y_len;i++)
	clust[cl].ymean[i] /= (double)clust[cl].nb_points;
 
 
  /* compute all covariance-matrices Cxx */
  for (l=0;l<nb_points;l++)
    {
      cl=clustered_data[l];
      for (i=0;i<x_len;i++)
	for (j=0;j<x_len;j++)
	  clust[cl].R[i][j] += ((train->data[l][i]-clust[cl].xmean[i]) * (train->data[l][j]-clust[cl].xmean[j]));
    }

  for (cl=0;cl<nb_clusters;cl++)
    if (clust[cl].nb_points>0)
      {
	nbpts_dbl=(double)clust[cl].nb_points;
	for (i=0;i<x_len;i++)
	  for (j=0;j<x_len;j++)
	    clust[cl].R[i][j] = clust[cl].R[i][j]/nbpts_dbl;
      }

      
  /* cross cor */
  for (l=0;l<nb_points;l++)
    {
      cl=clustered_data[l];
      for (i=0;i<x_len;i++)
	for (j=0;j<y_len;j++)
	  clust[cl].vect[i][j] += ((train->data[l][i]-clust[cl].xmean[i]) * (train->data[l][j+x_len]-clust[cl].ymean[j]));
    }
  for (cl=0;cl<nb_clusters;cl++)
    if (clust[cl].nb_points>0)
      {
	nbpts_dbl=(double)clust[cl].nb_points;
	for (i=0;i<x_len;i++)
	  for (j=0;j<y_len;j++)
	    clust[cl].vect[i][j] = clust[cl].vect[i][j]/nbpts_dbl;
      }
  
  /* Cyy */
  for (l=0;l<nb_points;l++)
    {
      cl=clustered_data[l];
      for (i=0;i<y_len;i++)
	for (j=0;j<y_len;j++)
	  clust[cl].Cyy[i][j] += ((train->data[l][i+x_len]-clust[cl].ymean[i]) * (train->data[l][j+x_len]-clust[cl].ymean[j]));
    }

  for (cl=0;cl<nb_clusters;cl++)
    if (clust[cl].nb_points>0)
      {
	nbpts_dbl=(double)clust[cl].nb_points;
	for (i=0;i<y_len;i++)
	  for (j=0;j<y_len;j++)
	    clust[cl].Cyy[i][j] = clust[cl].Cyy[i][j]/nbpts_dbl;
      }


  free((void*)clustered_data);


}

void compute_clust_filter(struct node *clust, struct data *train)
{
  int x_len,y_len;
  double **Rinv;
  int i,j,m;
  double nbpts_dbl;
  int k;


  if (clust->nb_points <=0)
    {
      fprintf(stderr,"Called compute_clust_filter for empty cluster. Shouldn't happen. nb_points=%d\n\a\n",clust->nb_points);
      exit(-1);
    }

  x_len = train->x_len;
  y_len = train->y_len;
  nbpts_dbl = (double)clust->nb_points;

  Rinv=(double**)multialloc(sizeof(double),2,x_len,x_len);



  for (i=0;i<x_len;i++)
    for (j=0;j<x_len;j++)
      Rinv[i][j] = clust->R[i][j];



#ifdef PINV
  pseudoinvert(Rinv,x_len);
#else 
  /* ADD Inversion constant to avoid singularity */
  for (i=0;i<x_len;i++)
    Rinv[i][i] += INVCONST;
 
  invert2(Rinv, x_len);
#endif


 

  for (i=0;i<x_len;i++)
    for (j=0;j<y_len;j++)
      clust->a[i][j] = (double)0.0;


  /* compute a = Cxx^-1 * Cxy , i.e. cov instead of corr. matrices*/
  for (i=0;i<x_len;i++)
    for (j=0;j<y_len;j++)
      for (m=0;m<x_len;m++)
	clust->a[i][j] += (Rinv[i][m] * clust->vect[m][j]);

  /* Compute b = mean_y - a * mean_x */
  for (i=0;i<y_len;i++)
    clust->b[i] = clust->ymean[i];

  for (j=0;j<y_len;j++)
    for (k=0;k<x_len;k++)
      clust->b[j] -= (clust->xmean[k] * clust->a[k][j]);



  multifree((void**)Rinv,2);
}




void compute_clust_prederr(struct node *clust, struct data *train)
{
  /* Since I'm adding a small diagonal constant to R=x'x, */
  /* orthogonality doesn't hold exactly. Thus computing the */
  /* mse from the 2nd order statistics involves 3 terms */
  /* i.e. need one term yhat'yhat, thus need Q=a*a' */

  /* mse=sum_s{y_s'*y_s - 2x_s'*A'*y_s + x_s'(A'A)x_s} */

  /* where yhat_s=A*x_s, i.e. A=clust->a'           */
 

  int x_len,y_len;
  double mse;
  int i,j;
  double **tmp,**a_t,**mat1,**mat2,**fin_mat;
  int rows,cols;


  if (clust->nb_points <=0)
    {
      fprintf(stderr,"Called compute_clust_pred_err for empty cluster. Shouldn't happen. nb_points=%d\n\a\n",clust->nb_points);
      exit(-1);
    }

  x_len = train->x_len;
  y_len = train->y_len;


 
  /* pred_err = nb_points * trace(Cyy - 2a^t * Cxy + a^t * Cxx * a)   
     
     mat1 = a^t * Cxy
     mat2 = a^t * Cxx * a
     
     Note: mat1 might not be exactly symmetric, so one should really
     use  a^t * Cxy + Cxy^t * a , however, doesn't make a difference 
     for the trace.
  */

  
  /* *** COMPUTE mat1 ********/
  a_t=matrix_trans(clust->a, x_len, y_len);

  mat1 = matrix_prod(a_t, y_len, x_len, clust->vect, x_len, y_len,&rows, &cols);
  if ((rows!=y_len)||(cols!=y_len))
    {
      fprintf(stderr,"Error in compute clust_prederr\n\a\n");
      exit(-1);
    }

  /* compute mat2 ************/
  tmp=matrix_prod(clust->R, x_len, x_len, clust->a, x_len, y_len,&rows, &cols);
  if ((rows!=x_len)||(cols!=y_len))
    {
      fprintf(stderr,"Error in compute clust_prederr\n\a\n");
      exit(-1);
    }
  
  mat2=matrix_prod(a_t, y_len, x_len, tmp, x_len, y_len,&rows, &cols);
  if ((rows!=y_len)||(cols!=y_len))
    {
      fprintf(stderr,"Error in compute clust_prederr\n\a\n");
      exit(-1);
    }

  /* now compute mat_err = (Cyy - 2a^t * Cxy + a^t * Cxx * a) 
                         = Cyy - 2mat1 + mat2 */
  
  fin_mat=(double**)multialloc(sizeof(double),2,y_len,y_len);
  for (i=0;i<y_len;i++)
    for (j=0;j<y_len;j++)
      fin_mat[i][j] = clust->Cyy[i][j] - 2.0*mat1[i][j] + mat2[i][j];
  
  mse=matrix_trace(fin_mat,y_len,y_len);
  mse = mse * (double)clust->nb_points;


  multifree((void**)a_t,2);
  multifree((void**)mat1,2);
  multifree((void**)tmp,2);
  multifree((void**)mat2,2);
  multifree((void**)fin_mat,2);

  /*printf("mse=%g\n",mse);*/
  clust->pred_err = mse;

}










void check_clust_prederr(struct node *clust,struct data *train, struct cmds *params)
{
  int nb_points,nb_clusters,x_len, y_len;
  double *y_hat;
  int i,j,cl,k;
  int stat=0;
  double er,dif;
  double *mse;



  nb_clusters = params->nb_clusters;
  nb_points=train->nb_points;
  x_len = train->x_len; 
  y_len = train->y_len; 
  
  y_hat = (double*)malloc(sizeof(double)*y_len);
  mse = (double*)malloc(sizeof(double)*nb_clusters);
  if ((y_hat==NULL) || (mse==NULL))
    {
      fprintf(stderr,"Can't alloc y_hat.\a\n");
      exit(-1);
    }
  
  for (i=0;i<nb_clusters;i++)
    mse[i]=(double)0.0;

  for (i=0;i<nb_points;i++)
    {
      for (j=0;j<y_len;j++)
	y_hat[j]=0.0;
      
      cl=determine_cluster(train->data[i], x_len, clust, params);


      for (j=0;j<y_len;j++)
	for (k=0;k<x_len;k++)
	  y_hat[j] += (train->data[i][k] * clust[cl].a[k][j]);




      for (j=0;j<y_len;j++)
	y_hat[j] += clust[cl].b[j];


      er=0.0;

      for (j=0;j<y_len;j++)
	er += ((y_hat[j] - train->data[i][j+x_len])* (y_hat[j] - train->data[i][j+x_len]));


      mse[cl] += er;

    }

  for (cl=0;cl<nb_clusters;cl++)
    {
      dif=fabs(mse[cl]-clust[cl].pred_err);
      if (mse[cl]>1e-8)
	{
	  if (dif/mse[cl] > 0.001)
	    {
	      printf("Error: cl=%d dif=%g mse=%g pred_err=%g \a\n",cl,dif,mse[cl],clust[cl].pred_err);
	      stat=1;
	    }
	}
      else
	if (dif > 1e-3)
	  {
	    printf("Error: cl=%d dif=%g mse=%g pred_err=%g \a\n",cl,dif,mse[cl],clust[cl].pred_err);
	    stat=1;
	  }

      if (stat==1)
	{
	  writeMatrix("a",clust[cl].a,x_len,y_len);
	  writeVector("b",clust[cl].b,y_len);
	  writeVector("xmean",clust[cl].xmean,x_len);
	  writeVector("ymean",clust[cl].ymean,y_len);
	  writeMatrix("vect",clust[cl].vect,x_len,y_len);

	  writeMatrix("R",clust[cl].R,x_len,x_len);


	  exit(-1);


	}



    }


  free((void*)y_hat);
  free((void*)mse);

}
 
