#include "../include/agt_io.h"

#define TRAIN 0
#define PRUNE 1

/* ********************************************************/
/* ****** I/O routines ************************************/
/* ********************************************************/

void read_trainfile(struct data *train, char *fname)
{
  int nb_points,x_len,y_len;
  int stat;
  int i,j;
  double dummy;
  char fname_sz[100];
  
  /* open file */
  FILE *fp, *fp_sz;

  
  
  /* read header */
  sprintf(fname_sz,"%s_sz",fname);
  fp_sz=fopen(fname_sz,"r");
  if (fp_sz==NULL)
    {
      fprintf(stderr,"Error: Can't open size-file %s\a\n\n",fname_sz);
      exit(-1);
    }
  stat=fscanf(fp_sz,"%d %d %d",&nb_points,&x_len,&y_len);
  if (stat!=3)
    {
      fprintf(stderr,"Error: Can't read size description from %s\a\n\n",fname_sz);
      exit(-1);
    }
  fclose(fp_sz);
  
  

  /* Open data files ********/
  fp=fopen(fname,"r");
  if (fp==NULL)
    {
      fprintf(stderr,"Error: Can't open trainfile %s\a\n\n",fname);
      exit(-1);
    }
 
  train->nb_points=nb_points;
  train->x_len=x_len;
  train->y_len=y_len;

  /* allocate data */
  train->data = (double**)multialloc(sizeof(double),2,nb_points,x_len+y_len);


 
  

  /* read data vectors */
  for (i=0;i<nb_points;i++)
    {
      for (j=0;j<x_len+y_len;j++)
	{
	  stat=fscanf(fp,"%lf",&train->data[i][j]);
	  if (stat!=1)
	    {
	      fprintf(stderr,"Error reading trainfile %s: stat=%d\a\n\n",fname,stat);
	      exit(-1);
	    }
	}

    } /* end for i */

  /* check if file ended */
  stat=fscanf(fp,"%lf",&dummy);
  if (stat==1)
    {
      fprintf(stderr,"WARNING: Training file too long. stat=%d, dummy=%f\n\a\n",stat,dummy);
    }

  fclose(fp);
}




void write_tree(struct node *clust,struct data *train,struct cmds *params)
{
  FILE *fp;
  int x_len,y_len,nb_clusters,nb_nodes;
  int k,i,j;
  
  
  x_len = train->x_len;
  y_len = train->y_len;
  nb_clusters = params->nb_clusters;
  nb_nodes = params->nb_nodes;


  fp=fopen(params->outfile,"w");
  if (fp==NULL)
    {
      fprintf(stderr,"Can't open output file %s\a\n",params->outfile);
      exit(-1);
    }

 
  fprintf(fp,"%d\n%d\n%d\n%d\n",nb_clusters,nb_nodes,x_len,y_len);
  
 
   
  for (k=0;k<nb_nodes;k++)
    {
      fprintf(fp,"%d ",clust[k].parent);
      for (j=0; j<x_len; j++)
	fprintf(fp,"%1.20g ",clust[k].mean[j]);
      fprintf(fp,"\n");
      for (i=0; i<x_len; i++)
	{
	  for (j=0; j<y_len; j++)
	    fprintf(fp,"%1.20g ",clust[k].a[i][j]);
	  fprintf(fp,"\n");
	}

      for (j=0; j<y_len; j++)
	fprintf(fp,"%1.20g ",clust[k].b[j]);
      fprintf(fp,"\n");

      for (i=0; i<y_len; i++)
	{
	  for (j=0; j<y_len; j++)
	    fprintf(fp,"%1.20g ",clust[k].yer2[i][j]);
	  fprintf(fp,"\n");
	}

      
      fprintf(fp,"%d\n",clust[k].prune_FLAG);
       
    }

  fclose(fp);
}






struct node* read_tree(char *fname, int *Xlen, int *Ylen, int *nbclust, int *nbnode)
{
  FILE *fp;
  int x_len,y_len,nb_clusters,nb_nodes;
  int k,i,j,stat;
  struct node *clust;
  double dummy;

  printf("Reading cluster info from %s\n",fname);

  fp=fopen(fname,"r");
  if (fp==NULL)
    {
      fprintf(stderr,"Can't open input cluster file %s\a\n",fname);
      exit(-1);
    }

 
  /* read parameters */
  stat=fscanf(fp,"%d %d %d %d",&nb_clusters,&nb_nodes,&x_len,&y_len);
  if (stat!=4)
    {
      fprintf(stderr,"Can't read from cluster file %s\a\n",fname);
      exit(-1);
    }

  *Xlen=x_len;
  *Ylen=y_len;
  *nbclust=nb_clusters;
  *nbnode=nb_nodes;
  


  /* Allocate clusters accordingly */
  /* allocate array of clusters including mean vector for each cluster*/
  clust=allocClusters(nb_nodes, x_len, y_len);


  /* READ clusters ***************************/
  for (k=0;k<nb_nodes;k++)
    {
      /* read tree parent of this node */
      stat=fscanf(fp,"%d",&clust[k].parent);
      if (stat!=1)
	{
	  fprintf(stderr,"error reading cluster mean\a\n");
	  exit(-1);
	} 


      /* Read mean vector for cluster k */
      for (j=0; j<x_len; j++)
	{
	  stat=fscanf(fp,"%lg",&clust[k].mean[j]);
	  if (stat!=1)
	    {
	      fprintf(stderr,"error reading cluster mean\a\n");
	      exit(-1);
	    } 
	}

      /* read filter for cluster k */
      for (i=0; i<x_len; i++)
	for (j=0; j<y_len; j++)
	  {	
	    stat=fscanf(fp,"%lg",&clust[k].a[i][j]);
	    if (stat!=1)
	      {
		fprintf(stderr,"error reading cluster filter (a) \a\n");
		exit(-1);
	      } 
	  }
      for (j=0; j<y_len; j++)
	{
	  stat=fscanf(fp,"%lg",&clust[k].b[j]);
	  if (stat!=1)
	    {
	      fprintf(stderr,"error reading cluster filter (b) \a\n");
	      exit(-1);
	    } 
	}

      for (i=0; i<y_len; i++)
	for (j=0; j<y_len; j++)
	  {
	    stat=fscanf(fp,"%lg",&clust[k].yer2[i][j]);
	    if (stat!=1)
	      {
		fprintf(stderr,"error reading cluster yer2 \a\n");
		exit(-1);
	      } 
	  }

      
      stat=fscanf(fp,"%d",&clust[k].prune_FLAG);
      if (stat!=1)
	{
	  fprintf(stderr,"error reading cluster prune_FLAG \a\n");
	  exit(-1);
	} 


   

    } /* end for k */
  
  /* check if file not too long */
  if (fscanf(fp,"%lg",&dummy) == 1)
    {
      fprintf(stderr,"cluster file too long??\a\n");
      exit(-1);
    }
  
  fclose(fp);

  return(clust);

}






void writeMatrix(char *fname, double **y_hat, int N, int M)
{
  FILE *fp;
  int i,j;

  fp=fopen(fname,"w");
  if (fp==NULL)
    {
      fprintf(stderr,"Can't open file %s\a\n",fname);
      exit(-1);
    }

  for (i=0;i<N;i++)
    {
      for (j=0;j<M;j++)
	fprintf(fp,"%1.20g ",y_hat[i][j]);
      fprintf(fp,"\n");
    }

  fclose(fp);
}

void writeVector(char *fname, double *y_hat, int N)
{
  FILE *fp;
  int j;

  fp=fopen(fname,"w");
  if (fp==NULL)
    {
      fprintf(stderr,"Can't open file %s\a\n",fname);
      exit(-1);
    }


  for (j=0;j<N;j++)
    fprintf(fp,"%1.20g ",y_hat[j]);


  fclose(fp);
}










