#include "../include/agt_main.h"


void make_agtree(struct cmds *params)
{
  struct data train,prune_set;
  struct node *clust;
  int i;
  int nbnods;
  double err_tr,err_pr;

  params->avg_cost_md = 0.0;
  params->avg_cost_tot = 0.0;

  params->nb_nodes = params->nb_clusters;

  readseed();

  
  /* READ BOTH TRAIN AND PRUNE DATA SETS */
  printf("reading data...");fflush(stdout);
  
  /* read trainfile */
  read_trainfile(&train, params->trainfile); 

  /* read prunefile, check vector dims */ 
  if (params->prune==1)
    {
      read_trainfile(&prune_set, params->prunefile); 
      if ((train.x_len != prune_set.x_len) || (train.y_len != prune_set.y_len))
	{
	  fprintf(stderr,"Error: Train and Prune file vector length's incompatible\n\a\n");
	  exit(-1);
	}
    }

  printf("done.\n");fflush(stdout);


  /* PERFORM INITIAL VQ TO OBTAIN LEAVES OF TREE
     USING TRAINING SET */
  printf("Performing initial vector quantization\n");
  clust=vq(&train, params);



  /* NOW COMPUTE MMSE LINEAR NODE FILTERS */
  compute_filters(clust, &train, params);




  /* Merge clusters into a binary tree */
  if (params->merge_limit != -1)
    {
      printf("\nMerging clusters into binary tree:\n");
      Merging(clust, &train, params);
    }


  
  /* if PRUNING is switched on, prune tree */
  if (params->prune==1)
    {
      for (i=0;i<params->nb_nodes;i++)
	clust[i].prune_FLAG=0;
      err_tr=compute_set_prederr(clust, &train, params);
      err_pr=compute_set_prederr(clust, &prune_set, params);
      /*printf("Error initial leaves , err_tr=%1.5g  err_pr=%1.5g  totalerr=%1.5g\n\n",err_tr,err_pr,err_tr+err_pr);*/
   
      printf("Computing errors for pruning data set\n");
      compute_all_prederrs_man(clust, &prune_set, params);
      
      printf("Pruning tree:\n");
      prune(clust, &prune_set, params);

      err_tr=compute_set_prederr(clust, &train, params);
      err_pr=compute_set_prederr(clust, &prune_set, params);
      /*printf("Err After pruning  , err_tr=%1.5g  err_pr=%1.5g  totalerr=%1.5g\n\n",err_tr,err_pr,err_tr+err_pr);*/


      nbnods=calc_nb_nodes(clust, params);
      fflush(stdout);fflush(stderr);

    } 

  /* Calculate variance of MMSE node predictors*/
  printf("Calculating prediction variances for all nodes\n");
  fflush(stdout);fflush(stderr);

  if (params->prune==1)
    CalcClustVar(clust,&prune_set, params);
  else
    CalcClustVar(clust,&train, params);
 

  /* Added 8/12/2000 : Compute expected conditional variances 
     and write to disk */
  printf("Computing expected conditional variances\n");
  compute_and_write_E_cond_var(clust,params,train.y_len);


  /* write output */
  write_tree(clust,&train,params);
 

  /* Free data */
  freeClusters(clust, 2*params->nb_clusters);
  multifree((void**)train.data,2);

  if (params->prune==1)
    multifree((void**)prune_set.data,2);
   
  
  writeseed();
  
  printf("Tree successfully built.\n\n");
  fflush(stderr);fflush(stdout);
}






/* ********************************************************/
/* ****** Routine for initial clustering      ************/
/* ********************************************************/


struct node* vq(struct data *train, struct cmds *params)
{
  int nb_points;
  int x_len,y_len;
  struct node *clust;
  int k;
  int nb_iter=5;
  
  nb_points=train->nb_points;
  x_len = train->x_len;
  y_len = train->y_len;

  /* check if train set has enough distinct data points
     to init all clusters with data points that have different values
     This takes a little time, but otherwise, the routine
     initClmeans below may loop forever if too many clusters
     are specified for a small training set.
  */
  printf("testing size of training set...");fflush(stdout);
  /* first check number of points overall*/
  if (nb_points<2*params->nb_clusters)
    {
      fprintf(stderr,"Error: Training set too small for specified number of initial clusters (=leaves). Number of training points=%d is less than 2*nb_leaves=%d. Reduce number of clusters or increase size of training set.\n\a\n",nb_points,2*params->nb_clusters);
      exit(-1);
    }

  

  test_trset_size(train, params);
  printf("done.\n");fflush(stdout);

  /* Allocate and init clusters */
  clust=allocClusters(2*params->nb_clusters, x_len, y_len);



  /* Init Cluster means */
  printf("initializing cluster means...");fflush(stdout);
  initClmeans(clust, train, params);
  printf("done.\n");fflush(stdout);
 
  classifyData(clust,train,params,nb_iter);
  


  for (k=0;k<params->nb_clusters;k++)
    printf("cluster %d, nb_points=%d\n",k,clust[k].nb_points);


  return(clust);
}
 



