#include "../include/agt_prune.h"



void prune(struct node *clust, struct data *train, struct cmds *params)
{
  int i,root;
  double tree_err;

  /* 1. clear prune_FLAG */
  for (i=0;i<params->nb_nodes;i++)
    clust[i].prune_FLAG = 0;

  /* 2. find tree_root */
  /* of course this assumes that we have single tree, not forest, i.e
     merge_limit=1
  */
  root=params->nb_nodes -1;
  if (clust[root].root_FLAG==0)
    {
      fprintf(stderr,"\n\nError: Last node not root node\a\n\n");
      exit(-1);
    }
  /* check tree, i.e. if paths from every leaf to root exist
     and if children are correctly assigned 
  */
  printf("Checking tree integrity before pruning...");fflush(stdout);
  check_full_tree(clust, params);
  printf("done.\n");fflush(stdout);



  /* now recursively prune nodes starting at root */
  tree_err=prune_node(root, clust,train,params);

  printf("total error after pruning: %g\n",tree_err);
}






double prune_node(int node, struct node *clust, struct data *train, struct cmds *params)
{
  double subtree_err,err_left,err_right;


  /* if we have a leaf, then subtree_error is leaf pred_err, return*/
  if ((clust[node].child1 == -1) || (clust[node].child2 == -1))
    {
      /* (obsolete) tree-check */
      if (clust[node].child1 != clust[node].child2)
	{
	  fprintf(stderr,"tree error in prune_node\n\a\n");
	  exit(-1);
	}
      printf("Considering leaf node %d\n",node);
      subtree_err = clust[node].pred_err;
    }
  else
    {
      err_left  = prune_node(clust[node].child1, clust,train,params);
      err_right = prune_node(clust[node].child2, clust,train,params);
      printf("node=%d pred_err=%g  left_err=%g   right_err=%g\n",node,clust[node].pred_err,err_left,err_right);


      if ((clust[node].pred_err<1e-9) || ((clust[node].pred_err <= err_left+err_right)))
	{
	  printf("Pruning nodes %d, %d into %d\n",clust[node].child1,clust[node].child2, node);
	  subtree_err = clust[node].pred_err;
	  /* Now go down both subtrees to leaves and mark everything
	     pruned 
	     LATER: Change code s.t. if encounter node in subtree
	     that is already marked as pruned, then stop marking
	  */
	  mark_subtree_pruned(clust[node].child1, clust);
	  mark_subtree_pruned(clust[node].child2, clust);
	}
      else
	{
	  printf("Keeping subtrees\n");
	  subtree_err = err_left + err_right;
	}
    }





  return(subtree_err);

}




void mark_subtree_pruned(int node, struct node *clust)
{
  clust[node].prune_FLAG = 1;
  if (clust[node].child1 != -1)
    {
      mark_subtree_pruned(clust[node].child1, clust);
      mark_subtree_pruned(clust[node].child2, clust);
    }
}






double compute_set_prederr(struct node *clust, struct data *train, struct cmds *params)
{
  int nb_points,nb_clusters,x_len, y_len;
  double y_hat_j;
  int i,j,cl,k;
  double pred_err=0.0;

  nb_clusters = params->nb_clusters;
  nb_points=train->nb_points;
  x_len = train->x_len; 
  y_len = train->y_len; 
  
  for (i=0;i<nb_points;i++)
    {
      cl=classify_xvec(train->data[i], x_len, clust, params);
      for (j=0;j<y_len;j++)
	{
	  y_hat_j = 0.0;
	  for (k=0;k<x_len;k++)
	    y_hat_j += (train->data[i][k] * clust[cl].a[k][j]);
	  

	  y_hat_j += clust[cl].b[j];

	  /* Now add to pred_err */
	  pred_err += ((y_hat_j - train->data[i][j+x_len])*(y_hat_j - train->data[i][j+x_len]));
	}
    }

  return(pred_err);
}
