#include "../include/agt_tree.h"

/* determine_node: routine for run-time prediction using tree */
/* first determines cluster=leaf and then runs up the tree */
/* to the root node corresponding to the cluster */
int determine_node(double *xvec, int x_len, struct node *clust, struct cmds *params)
{
  int node;

  /* get leaf cluster */
  node=determine_cluster(xvec, x_len, clust, params);
  
  /* now run up the tree */
  while (clust[node].parent != -1)
    node=clust[node].parent;

  return(node);
}


/* routine for pruned tree */
int classify_xvec(double *xvec, int x_len, struct node *clust, struct cmds *params)
{
  int node;

  /* get leaf cluster */
  node=determine_cluster(xvec, x_len, clust, params);
  
  if (clust[node].prune_FLAG==-1)
    {
      /* this is not a pruned tree */
      while (clust[node].parent != -1)
	node=clust[node].parent;
    }
  else
    {
      /* now run up the tree */
      while ((clust[node].parent != -1) && (clust[node].prune_FLAG==1))
	node=clust[node].parent;
    }

  return(node);
}







int calc_nb_nodes(struct node *clust, struct cmds *params)
{
  int ct,stat,i,j,node;
  int *nods;
  int ct2;


  nods=(int*)malloc(sizeof(int)*params->nb_nodes);
  if (nods==NULL)
    {
      fprintf(stderr,"malloc errror in calc_nb_nods, agt_tree.c\n\a\n");
      exit(-1);
    }     
 

  ct=0;
  for (i=0;i<params->nb_clusters;i++)
    {
      node=i;
      if (clust[node].prune_FLAG==-1)
	{
	  /* this is not a pruned tree */
	  while (clust[node].parent != -1)
	    node=clust[node].parent;
	}
      else
	{
	  /* now run up the tree */
	  while ((clust[node].parent != -1) && (clust[node].prune_FLAG==1))
	    node=clust[node].parent;
	}


      if (i==0)
	{
	  nods[ct]=node;
	  ct++;
	}
      else
	{
	  stat=0;
	  for (j=0;j<ct;j++)
	    {
	      if (nods[j]==node)
		{
		  stat=1;
		  break;
		}
	    }
	  if (stat==0)
	    {
	      nods[ct]=node;
	      ct++;
	    }
	}
    }

  printf("After pruning have a total of %d nodes.\n\n",ct);
  printf("Those nodes are:\n");
  for (i=0;i<ct;i++)
    printf("%d\n",nods[i]);
  printf("\n\n");


  ct2=0;
  for (i=0;i<params->nb_nodes;i++)
    {
      if (clust[i].prune_FLAG==0) 
	if ((clust[i].child1 == -1) || (clust[clust[i].child1].prune_FLAG==1))
	  {
	    if ((clust[i].parent!=-1) && (clust[clust[i].parent].prune_FLAG!=0))
	      {
		fprintf(stderr,"error: node %d, prune_FLAG=0 but parent_prune_FLAG!=0\n\a\n",i);
		exit(-1);
	      }
	  
	    if ((clust[i].child2!=-1) && (clust[clust[i].child2].prune_FLAG!=1))
	      {
		fprintf(stderr,"error: node %d, prune_FLAG=0, child1.prune_FLAG==1 but child2_prune_FLAG!=1\n\a\n",i);
		exit(-1);
	      }

	    ct2++;
	  }
    }

  
  if (ct!=ct2)
    {
      	fprintf(stderr,"error: After pruning, calculated different number of root nodes using 2 different methods, ct=%d, ct2=%d\n\a\n",ct,ct2);
	exit(-1);
    }
 

  free((void*)nods);


  return(ct);

}





void check_full_tree(struct node *clust, struct cmds *params)
{
  int i,rt,rt_old;

  for (i=0;i<params->nb_clusters;i++)
    {
      /* first check whether leaf has leaf properties */
      if ((clust[i].child1 != -1) || (clust[i].child2 != -1))
	{
	  fprintf(stderr,"leaf %d: one child != -1, ch1=%d ch2=%d\n\a\n",i,clust[i].child1,clust[i].child2);
	  exit(-1);
	}
     
      /* Now run up tree to root and check parent/child relationships
	 on the way 
      */
      rt=i;
      while (clust[rt].parent != -1)
	{
	  if (clust[i].root_FLAG != 0) /* since parent != -1, can't be root*/
	    {
	      fprintf(stderr,"leaf %d: root_FLAG != 0 !! Error: parent !=-1 but root_FLAG != 0",i);
	      exit(-1);
	    }

	  /* move tree one up */
	  rt_old=rt;
	  rt=clust[rt].parent;
	  
	  /* now check whether one of the new childdren is the node 
	     that we just move up from 
	  */
	  if ((clust[rt].child1 != rt_old) && (clust[rt].child2 != rt_old))
	    {
	      fprintf(stderr,"went up child->parent but parent.child1 and child2 != child\n\a\n");
	      exit(-1);
	    }
	  /* also, since we moved one up, this can't be a leaf. Thus
	     child1 must not equal child2
	  */
	  if (clust[rt].child1 == clust[rt].child2)
	    {
	      fprintf(stderr,"both children are the same\n\a\n");
	      exit(-1);
	    }

	  /* every non-leaf node must have 2 children */
	  if ((clust[rt].child1 == -1) || (clust[rt].child2==-1))
	    {
	      fprintf(stderr,"non leaf node has one child==-1\n\a\n");
	      exit(-1);
	    }

	  /* both children's parents must be this node */
	  if ((clust[clust[rt].child1].parent != rt) || (clust[clust[rt].child2].parent != rt))
	    {
	      fprintf(stderr,"some parent/child links broken\n\a\n");
	      exit(-1);
	    }
	  

	  
	} /* end while */

      /* so are we at the root now ?? */
      if (rt != params->nb_nodes-1)
	{
	  fprintf(stderr,"Tree error!! Path doesn't lead up to root\n\a\n");
	  exit(-1);
	}

    } /* end for i, loop over leafs*/

  /* root should have root_FLAG ==1 */
  if (clust[params->nb_nodes-1].root_FLAG != 1)
    {
      fprintf(stderr,"root's root_FLAG = %d \n\a\n",clust[params->nb_nodes-1].root_FLAG);
      exit(-1);
    }

}










void printTree(struct node *clust, struct cmds *params)
{
 int rt;

 rt=params->nb_nodes-1;
 if (clust[rt].root_FLAG==0)
   {
     fprintf(stderr,"\n\nError: Last node not root node\a\n\n");
     exit(-1);
   }
 printNode(clust[rt].child1,clust);

 printf("node=%d child1=%d child2=%d parent=%d err=%g nbpoints=%d prune_FLAG=%d\n",rt,clust[rt].child1,clust[rt].child2,  clust[rt].parent, clust[rt].pred_err, clust[rt].nb_points,clust[rt].prune_FLAG);

 printNode(clust[rt].child2,clust);



}


void printNode(int node,struct node* clust)
{
  if (clust[node].child1 != -1)
    printNode(clust[node].child1,clust);
 
  printf("node=%d child1=%d child2=%d parent=%d err=%g nbpoints=%d  prune_FLAG=%d\n",node,clust[node].child1,clust[node].child2,  clust[node].parent, clust[node].pred_err, clust[node].nb_points,clust[node].prune_FLAG);

  if (clust[node].child2 != -1)
    printNode(clust[node].child2,clust);
}

