#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include "nrutil.h"
#include "allocate.h"

static void ludcmp(double **a,int n,int *indx,double *d);
static void lubksb(double **a,int n,int *indx,double b[]);

void exit();


/* For invert 1 and 2 */
/*
main()
{
	int i,j;
	double aa[3][3],*a[3];
        void invert1(),invert2();

        a[0]=&(aa[0][0]);
        a[1]=&(aa[1][0]);
        a[2]=&(aa[2][0]);

	a[0][0]=1; a[0][1]=2; a[0][2]=3;
	a[1][0]=1; a[1][1]=2; a[1][2]=2;
	a[2][0]=2; a[2][1]=1; a[2][2]=1;

	invert2((double **)a,3);
	for(i=0; i<1000; i++) invert1((double **)a,3);
	printf("finished\n");

	for(i=0; i<3; i++) {
	  for(j=0; j<3; j++) printf("%lf ",a[i][j]);
          printf("\n");
        }

        return(0);
}
*/


/* For invert 3 and 4 */
/*
main()
{
	int i,j;
	double a[3][3];

	a[0][0]=1; a[0][1]=2; a[0][2]=3;
	a[1][0]=1; a[1][1]=2; a[1][2]=2;
	a[2][0]=2; a[2][1]=1; a[2][2]=1;

	invert((double *)a,3);
	for(i=0; i<1000; i++) invert((double *)a,3);
	printf("finished\n");

	for(i=0; i<3; i++) {
	  for(j=0; j<3; j++) printf("%lf ",a[i][j]);
          printf("\n");
        }

        return(0);
}
*/

/* answer = (
(0.0, -0.33333333333333337, 0.66666666666666663)
(-1.0, 1.66666666666666674, -0.333333333333333315)
(1.0, -1.0, 0.0)
)*/





/* inverts a matrix input as a 2D array. Requires n<=6 */
void invert1(
  double **a, /* input/output matrix */
  int	n)    /* dimension */
{
	int  i,j,indx[7];
	double  y[7][7],col[7],d;

	if(n>6) { printf("invert: n too large\n"); exit(-1); }

        /* Shift to index to a[1-n][1-n] */
	a--;
	for(i=1; i<=n; i++) a[i]--;

        /* Invert matrix */
	ludcmp(a,n,indx,&d);
	for(j=1; j<=n; j++) {
	  for(i=1; i<=n; i++) col[i]=0.0;
	  col[j]=1.0;
	  lubksb(a,n,indx,col);
	  for(i=1; i<=n; i++) y[i][j]=col[i];
	}

	for(i=1; i<=n; i++)
	for(j=1; j<=n; j++) a[i][j]=y[i][j];

        /* Shift to index back */
	for(i=1; i<=n; i++) a[i]++;
	a++;
}

/* inverts a matrix of arbitrary size input as a 2D array. */ 
void invert2(
  double **a, /* input/output matrix */
  int	n)    /* dimension */
{
	int  i,j,*indx;
	double  **y,*col,d,**dmatrix();

        /* Shift to index to a[1-n][1-n] */
	a--;
	for(i=1; i<=n; i++) a[i]--;

	indx = ivector(1,n);
	y = dmatrix(1,n,1,n); 
	col = (double *)mget_spc(n,sizeof(double)); col--;

	ludcmp(a,n,indx,&d);
	for(j=1; j<=n; j++) {
	  for(i=1; i<=n; i++) col[i]=0.0;
	  col[j]=1.0;
	  lubksb(a,n,indx,col);
	  for(i=1; i<=n; i++) y[i][j]=col[i];
	} 

	for(i=1; i<=n; i++)
	for(j=1; j<=n; j++) a[i][j]=y[i][j];

        /* Shift to index as a[1-n][1-n] */
	for(i=1; i<=n; i++) a[i]++;
	a++;

	free_ivector(indx,1,n);
	free_dmatrix(y,1,n,1,n);
	free((char *)(col+1));
}

/* inverts a matrix input as a 1D vector. Requires n<=6 */
void invert3(double *a,int n)
{
	int  i,j,indx[7];
	double  **aa,yy[7][7],col[7],d;

	if(n>6) { printf("invert: n too large\n"); exit(-1); }

	a--;
	aa = (double **)mget_spc(n,sizeof(double *)); aa--;
	for(i=1; i<=n; i++) aa[i]= a+(i-1)*n;

	ludcmp(aa,n,indx,&d);
	for(j=1; j<=n; j++) {
	  for(i=1; i<=n; i++) col[i]=0.0;
	  col[j]=1.0;
	  lubksb(aa,n,indx,col);
	  for(i=1; i<=n; i++) yy[i][j]=col[i];
	}

	for(i=1; i<=n; i++)
	for(j=1; j<=n; j++) aa[i][j]=yy[i][j];

	free((char *)(aa+1));
}


/* inverts a matrix of arbitrary size input as a 1D vector. */ 
void invert4(double *a,int n)
{
	int  i,j,*indx;
	double  **aa,**yy,*col,d,**dmatrix();

	a--;
	aa = (double **)mget_spc(n,sizeof(double *)); aa--;
	for(i=1; i<=n; i++) aa[i]= a+(i-1)*n;

	indx = ivector(1,n);
	yy = dmatrix(1,n,1,n);
	col = (double *)mget_spc(n,sizeof(double)); col--;

	ludcmp(aa,n,indx,&d);
	for(j=1; j<=n; j++) {
	  for(i=1; i<=n; i++) col[i]=0.0;
	  col[j]=1.0;
	  lubksb(aa,n,indx,col);
	  for(i=1; i<=n; i++) yy[i][j]=col[i];
	}

	for(i=1; i<=n; i++)
	for(j=1; j<=n; j++) aa[i][j]=yy[i][j];

	free_ivector(indx,1,n);
	free_dmatrix(yy,1,n,1,n);
	free((char *)(aa+1));
	free((char *)(col+1));
}


#define TINY 1.0e-20;

static void ludcmp(double **a,int n,int *indx,double *d)
{
	int i,imax=0,j,k;
	double big,dum,sum,temp;
	double *vv;

	vv=vector(1,n);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) nrerror("Singular matrix in routine LUDCMP");
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) a[i][j] *= dum;
		}
	}
	free_vector(vv,1,n);
}

#undef TINY

static void lubksb(double **a,int n,int *indx,double b[])
{
	int i,ii=0,ip,j;
	double sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}

