#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <math.h>
#define NRANSI
/*#include "nrutil.h"*/
/* 
   This file is taken from numerical recipes in C, but it is a NEWER version since the
   old one is unstable! Also, the original file in nrc is called 'dsvdcmp.c' (double arithmetic).
   Furthermore, I've copied in all routines needed from nr_util.h, since
   1) the new nrutil.h is not compatible with the old one which is needed by invert.c
     and eigen.c - (don't want to replace invert and eigen since I can't test the new ones right now)
   2) I didn't like the global defs in nrutil.h - especially since the unused ones result in lots
     of stupid compiler warnings.
   Thus I've renamed all defines and subroutines needed with _SVD to obtain a
   standalone svdcmp .
   (Tom Frese, 3/8/99)
   */
double dpythag_svd(double a, double b);
void free_dvector_svd(double *v, long nl, long nh);
double *dvector_svd(long nl, long nh);
void nrerror_svd(char error_text[]);

#define NR_END_SVD 1

static double dsqrarg_svd;
#define DSQR_SVD(a) ((dsqrarg_svd=(a)) == 0.0 ? 0.0 : dsqrarg_svd*dsqrarg_svd)

static double dmaxarg1_svd,dmaxarg2_svd;
#define DMAX_SVD(a,b) (dmaxarg1_svd=(a),dmaxarg2_svd=(b),(dmaxarg1_svd) > (dmaxarg2_svd) ?\
        (dmaxarg1_svd) : (dmaxarg2_svd))

static int iminarg1_svd,iminarg2_svd;
#define IMIN_SVD(a,b) (iminarg1_svd=(a),iminarg2_svd=(b),(iminarg1_svd) < (iminarg2_svd) ?\
        (iminarg1_svd) : (iminarg2_svd))

#define SIGN_SVD(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
#define FREE_ARG_SVD char*
 

void nrerror_svd(char error_text[])
/* Numerical Recipes standard error handler */
{
        fprintf(stderr,"Numerical Recipes run-time error...\n");
        fprintf(stderr,"%s\n",error_text);
        fprintf(stderr,"...now exiting to system...\n");
        exit(1);
}


double *dvector_svd(long nl, long nh)
/* allocate a double vector with subscript range v[nl..nh] */
{
        double *v;

        v=(double *)malloc((size_t) ((nh-nl+1+NR_END_SVD)*sizeof(double)));
        if (!v) nrerror_svd("allocation failure in dvector()");
        return v-nl+NR_END_SVD;
}


void free_dvector_svd(double *v, long nl, long nh)
/* free a double vector allocated with dvector() */
{
        free((FREE_ARG_SVD) (v+nl-NR_END_SVD));
}


double dpythag_svd(double a, double b)
{
        double absa,absb;
        absa=fabs(a);
        absb=fabs(b);
        if (absa > absb) return absa*sqrt(1.0+DSQR_SVD(absb/absa));
        else return (absb == 0.0 ? 0.0 : absb*sqrt(1.0+DSQR_SVD(absa/absb)));
}



void svdcmp(double **a, int m, int n, double w[], double **v)
{
  /*double dpythag(double a, double b);*/
	int flag,i,its,j,jj,k,l=0,nm=0;
	double anorm,c,f,g,h,s,scale,x,y,z,*rv1;

	rv1=dvector_svd(1,n);
	g=scale=anorm=0.0;
	for (i=1;i<=n;i++) {
		l=i+1;
		rv1[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m) {
			for (k=i;k<=m;k++) scale += fabs(a[k][i]);
			if (scale) {
				for (k=i;k<=m;k++) {
					a[k][i] /= scale;
					s += a[k][i]*a[k][i];
				}
				f=a[i][i];
				g = -SIGN_SVD(sqrt(s),f);
				h=f*g-s;
				a[i][i]=f-g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
					f=s/h;
					for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
				}
				for (k=i;k<=m;k++) a[k][i] *= scale;
			}
		}
		w[i]=scale *g;
		g=s=scale=0.0;
		if (i <= m && i != n) {
			for (k=l;k<=n;k++) scale += fabs(a[i][k]);
			if (scale) {
				for (k=l;k<=n;k++) {
					a[i][k] /= scale;
					s += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = -SIGN_SVD(sqrt(s),f);
				h=f*g-s;
				a[i][l]=f-g;
				for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
				for (j=l;j<=m;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
					for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
				}
				for (k=l;k<=n;k++) a[i][k] *= scale;
			}
		}
		anorm=DMAX_SVD(anorm,(fabs(w[i])+fabs(rv1[i])));
	}
	for (i=n;i>=1;i--) {
		if (i < n) {
			if (g) {
				for (j=l;j<=n;j++) v[j][i]=(a[i][j]/a[i][l])/g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
					for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
				}
			}
			for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
		}
		v[i][i]=1.0;
		g=rv1[i];
		l=i;
	}
	for (i=IMIN_SVD(m,n);i>=1;i--) {
		l=i+1;
		g=w[i];
		for (j=l;j<=n;j++) a[i][j]=0.0;
		if (g) {
			g=1.0/g;
			for (j=l;j<=n;j++) {
				for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
				f=(s/a[i][i])*g;
				for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
			}
			for (j=i;j<=m;j++) a[j][i] *= g;
		} else for (j=i;j<=m;j++) a[j][i]=0.0;
		++a[i][i];
	}
	for (k=n;k>=1;k--) {
		for (its=1;its<=30;its++) {
			flag=1;
			for (l=k;l>=1;l--) {
				nm=l-1;
				if ((double)(fabs(rv1[l])+anorm) == anorm) {
					flag=0;
					break;
				}
				if ((double)(fabs(w[nm])+anorm) == anorm) break;
			}
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l;i<=k;i++) {
					f=s*rv1[i];
					rv1[i]=c*rv1[i];
					if ((double)(fabs(f)+anorm) == anorm) break;
					g=w[i];
					h=dpythag_svd(f,g);
					w[i]=h;
					h=1.0/h;
					c=g*h;
					s = -f*h;
					for (j=1;j<=m;j++) {
						y=a[j][nm];
						z=a[j][i];
						a[j][nm]=y*c+z*s;
						a[j][i]=z*c-y*s;
					}
				}
			}
			z=w[k];
			if (l == k) {
				if (z < 0.0) {
					w[k] = -z;
					for (j=1;j<=n;j++) v[j][k] = -v[j][k];
				}
				break;
			}
			if (its == 30) nrerror_svd("no convergence in 30 dsvdcmp iterations");
			x=w[l];
			nm=k-1;
			y=w[nm];
			g=rv1[nm];
			h=rv1[k];
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
			g=dpythag_svd(f,1.0);
			f=((x-z)*(x+z)+h*((y/(f+SIGN_SVD(g,f)))-h))/x;
			c=s=1.0;
			for (j=l;j<=nm;j++) {
				i=j+1;
				g=rv1[i];
				y=w[i];
				h=s*g;
				g=c*g;
				z=dpythag_svd(f,h);
				rv1[j]=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g = g*c-x*s;
				h=y*s;
				y *= c;
				for (jj=1;jj<=n;jj++) {
					x=v[jj][j];
					z=v[jj][i];
					v[jj][j]=x*c+z*s;
					v[jj][i]=z*c-x*s;
				}
				z=dpythag_svd(f,h);
				w[j]=z;
				if (z) {
					z=1.0/z;
					c=f*z;
					s=h*z;
				}
				f=c*g+s*y;
				x=c*y-s*g;
				for (jj=1;jj<=m;jj++) {
					y=a[jj][j];
					z=a[jj][i];
					a[jj][j]=y*c+z*s;
					a[jj][i]=z*c-y*s;
				}
			}
			rv1[l]=0.0;
			rv1[k]=f;
			w[k]=x;
		}
	}
	free_dvector_svd(rv1,1,n);
}
#undef NRANSI
