#include "../include/interface_agtree.h"
 

void calc_prune_stats(struct geom_info *geom, int level, int *step, int *nb_train, int *nb_prune,int ***mask)
{
  int nb_coarse_points=0;
  
  int i,j;

  for (i=0;i<geom[level].n_y;i++)
    for (j=0;j<geom[level].n_x;j++)
      if (mask[level][i][j])
	nb_coarse_points++;
  
  *step = (int)((double)nb_coarse_points/10.0);
  if (*step < 5)
    *step = 5;
  
*step=2;


  *nb_prune= nb_coarse_points/(*step);
  *nb_train= nb_coarse_points - (*nb_prune);

  /*printf("total_points=%d prune_step=%d nb_train=%d nb_prune=%d\n",nb_coarse_points,*step,*nb_train,*nb_prune);*/


}


/* New version allows appending, depending on 'MODE' */
void wav_write_train_files(double ***X, double ****Z_true, int level, struct geom_info *geom, char *fnameTrain, char *fnamePrune, int MODE, int ***mask)
{
  int prune_step; 
  
  FILE *fpT,*fpP;
  FILE *fpTsz,*fpPsz; /* unnormalized data */
  char fmode[3];

  int x_cols,y_cols;
  int pt=0,nb_prune,nb_train;
  char fnamePsz[50],fnameTsz[50];
  int stat,old_nb_train,old_nb_prune,oldP_x_cols,oldT_x_cols,oldP_y_cols,oldT_y_cols;;


  x_cols=WINDOW_SIZE * WINDOW_SIZE - 1;
  y_cols=3;


  if (MODE==NEW)
    strcpy(fmode,"w");
  else 
    if (MODE==APPEND)
      {
	strcpy(fmode,"a");
      }
    else
      {
	fprintf(stderr,"Unknown mode in write_train_files_piecewise\n\a\n");
	exit(-1);
      }



  sprintf(fnameTsz,"%s_sz",fnameTrain);
  sprintf(fnamePsz,"%s_sz",fnamePrune);

  calc_prune_stats(geom, level, &prune_step, &nb_train, &nb_prune,mask);
  /* level is coarse scale level */



  /* open files */
  fpT=fopen(fnameTrain,fmode);
  fpP=fopen(fnamePrune,fmode);

  if ((fpT==NULL) || (fpP==NULL))
    {
      fprintf(stderr,"Error in write_train_files: Can't open files for writing\n");
      exit(-1);
    }
  

  if (MODE==APPEND)
    {
      fpTsz=fopen(fnameTsz,"r");
      fpPsz=fopen(fnamePsz,"r");
      if ((fpTsz==NULL) || (fpPsz==NULL))
	{
	  fprintf(stderr,"Can't open old size files %s and/or %s/\n\a\n",fnameTsz,fnamePsz);
	  exit(-1);
	}

      stat=fscanf(fpTsz,"%d %d %d",&old_nb_train, &oldT_x_cols, &oldT_y_cols);
      if (stat!=3)
	{
	  fprintf(stderr,"Can't read from file %s\n\a\n",fnameTsz);
	  exit(-1);
	}
      stat=fscanf(fpPsz,"%d %d %d",&old_nb_prune, &oldP_x_cols, &oldP_y_cols);
      if (stat!=3)
	{
	  fprintf(stderr,"Can't read from file %s\n\a\n",fnamePsz);
	  exit(-1);
	}
      
      if ((oldT_x_cols!=oldP_x_cols) || (oldT_y_cols!=oldP_y_cols) || (oldT_x_cols!=x_cols) || (oldT_y_cols!=y_cols))
	{
	  fprintf(stderr,"Mismatch x_cols/y_cols\n\a\n");
	  exit(-1);
	}

      fclose(fpTsz);
      fclose(fpPsz);
    }
  else
    {
      old_nb_train = 0;
      old_nb_prune = 0;
    }
  
  nb_train += old_nb_train;
  nb_prune += old_nb_prune;


  fpTsz=fopen(fnameTsz,"w");
  fpPsz=fopen(fnamePsz,"w");
  if ((fpTsz==NULL) || (fpPsz==NULL))
    {
      fprintf(stderr,"Can't open files %s and/or %s for writing\n\a\n",fnameTsz,fnamePsz);
      exit(-1);
    }
  
  fprintf(fpTsz,"%d\n%d\n%d\n",nb_train,x_cols,y_cols);
  fprintf(fpPsz,"%d\n%d\n%d\n",nb_prune,x_cols,y_cols);

  fclose(fpTsz);
  fclose(fpPsz);

  
  write_train_image(X[level], Z_true[level-1], geom[level].n_y, geom[level].n_x,fpT,fpP,prune_step,&pt,mask[level],level);

  fclose(fpT);
  fclose(fpP);


}

  
void write_train_image(double **Xcoarse_un, double ***Zfine_un, int N_coarse, int M_coarse, FILE *fpT, FILE *fpP, int prune_step, int *pt, int **coarse_mask,int level_coarse)
{
  double *out;
  int *ci_index,*cj_index;
  int k,ctr_i,ctr_j;
  double ctr_pixel_val;
  double **Xcoarse,***Zfine;
  int i,j;
  double cr_mean;
  double fct=1.0;
  int x_cols,y_cols;

  cr_mean=1.0;
  i=1;j=1;fct=1.0;
  

  x_cols=WINDOW_SIZE * WINDOW_SIZE - 1;
  y_cols=3;


  out=(double*)mget_spc(x_cols+y_cols,sizeof(double));
  ci_index=(int*)mget_spc(WINDOW_SIZE*WINDOW_SIZE,sizeof(int));
  cj_index=(int*)mget_spc(WINDOW_SIZE*WINDOW_SIZE,sizeof(int));
  

  
  fct=1.0;

  cr_mean=calc_corrected_mean(Xcoarse_un,coarse_mask,N_coarse, M_coarse,1);
  cr_mean/=fct;
  if (cr_mean<0.0)
    {
      fprintf(stderr,"Error: Corrected mean=%f < 0.0 !!\n\a\n",cr_mean);
      exit(-1);
    }
  
  Xcoarse=(double**)multialloc(sizeof(double),2,N_coarse,M_coarse);
  Zfine=(double***)multialloc(sizeof(double),3,3,N_coarse,M_coarse);
  
  for (i=0;i<N_coarse;i++)
    for (j=0;j<M_coarse;j++)
      Xcoarse[i][j]=Xcoarse_un[i][j]/cr_mean;
  
  for (k=0;k<3;k++)
    for (i=0;i<N_coarse;i++)
      for (j=0;j<M_coarse;j++)
	Zfine[k][i][j]=Zfine_un[k][i][j]/cr_mean;
    
  /*printf("Normalize train_lev: cr_mean=%f\n",cr_mean);*/


  
  for (ctr_i=0;ctr_i<N_coarse;ctr_i++)
    for (ctr_j=0;ctr_j<M_coarse;ctr_j++)
      if (coarse_mask[ctr_i][ctr_j])
	{
	  get_coarse_index(ci_index, cj_index, ctr_i, ctr_j, N_coarse, M_coarse);
	  ctr_pixel_val = Xcoarse[ci_index[x_cols/2]][cj_index[x_cols/2]];
	  /* x, causal part of coarse scale window */
	  for (k=0;k<x_cols/2;k++)
	    out[k]=Xcoarse[ci_index[k]][cj_index[k]]-ctr_pixel_val;
	  
	  /* x, anticausal part, after leaving out center pixel */
	  for (k=x_cols/2+1;k<=x_cols;k++)
	    out[k-1]=Xcoarse[ci_index[k]][cj_index[k]]-ctr_pixel_val;
	  
	  for (k=0;k<y_cols;k++)
	    out[k+x_cols]=Zfine[k][ctr_i][ctr_j];
	  
	  
	  if ((ctr_i!=ci_index[x_cols/2])||(ctr_j!=cj_index[x_cols/2]))
	    {
	      fprintf(stderr,"Indexing error in write_train_image\n\a\n");
	      exit(-1);
	    }
       
	  /* prune or train? */
	  if (*pt < prune_step-1)
	    {
	      /* TRAIN */
	      for (k=0;k<x_cols+y_cols;k++)
		fprintf(fpT,"%1.15g ",out[k]);
	      fprintf(fpT,"\n");
	      (*pt)++;
	    }
	  else
	    {	  
	      /* PRUNE */
	      for (k=0;k<x_cols+y_cols;k++)
		fprintf(fpP,"%1.15g ",out[k]);
	      fprintf(fpP,"\n");
	      *pt=0;	    
	    }
	}
  
  free((void*)out);
  free((void*)ci_index);
  free((void*)cj_index);
  multifree((void*)Zfine,3);
  multifree((void*)Xcoarse,2);
}




void get_coarse_index(int *i_index, int *j_index, int ctr_i, int ctr_j, int N_coarse, int M_coarse)
{
  int ct=0;
  int i,j,ii,jj;
  int offs;

  /* write coarse scale window */
  
  offs = (int)(WINDOW_SIZE/2);
  ct=0;
  


  for (i=ctr_i-offs;i<=ctr_i+offs;i++)
    for (j=ctr_j-offs;j<=ctr_j+offs;j++)
      {
	circ_boundary_conds(&ii,&jj,i,j,N_coarse,M_coarse);
	      
	i_index[ct]=ii;
	j_index[ct]=jj;
	ct++;
      }
}

void get_fine_index(int *i_index, int *j_index, int ctr_i, int ctr_j)
{  
  /* ctr_i, ctr_j are the COARSE scale center 
     coordinates of prediction window */
  int ct=0;
  int i,j;

  /* write fine scale 4x4 */
  for (i=2*ctr_i;i<=2*ctr_i+1;i++)
    for (j=2*ctr_j;j<=2*ctr_j+1;j++)
      {
	i_index[ct]=i;
	j_index[ct]=j;
	ct++;
      }
}




double wav_predict_fixvar(double ***Z, double ****Zvar, double **coarse_un, int N_coarse, int M_coarse,struct node *clust,int nb_clusters,int nb_nodes,int **cluster_map,int level_coarse, int **mask_coarse,int mask_FLAG)
{
  int ctr_i,ctr_j,k,j,cl,i;
  int X_len,Y_len;
  int *ci_index,*cj_index,fi_index[4],fj_index[4];
  double *X_data,Y_data[3],Y_var[3][3];
  struct cmds params;

  double ctr_pixel_val;
  double **coarse;
  double cr_mean=1.0;
  int NORM_FLAG=0;
 

  X_len = WINDOW_SIZE*WINDOW_SIZE-1;
  Y_len=3; /* cannot be changed w/o modifying code*/
  
  ci_index=(int*)mget_spc(WINDOW_SIZE*WINDOW_SIZE,sizeof(int));
  cj_index=(int*)mget_spc(WINDOW_SIZE*WINDOW_SIZE,sizeof(int));
  X_data=(double*)mget_spc(X_len,sizeof(double));
  

  /* FIRST NORMALIZE COARSE SCALE INPUT */
  cr_mean=calc_corrected_mean(coarse_un,mask_coarse, N_coarse, M_coarse,mask_FLAG);


  /*printf("Final cr_mean=%f\n",cr_mean);*/
  if (cr_mean<0.0)
    {
      fprintf(stderr,"Corrected mean=%f < 0.0 !!\n\a\n",cr_mean);
      exit(-1);
    }

  if (cr_mean>1e-15)
    {
      NORM_FLAG=1;
      coarse=(double**)multialloc(sizeof(double),2,N_coarse,M_coarse);
      for (i=0;i<N_coarse;i++)
	for (j=0;j<M_coarse;j++)
	  coarse[i][j]=coarse_un[i][j]/cr_mean;
    }
  else
    {
      NORM_FLAG=0;
      coarse=coarse_un;
    }





  params.nb_clusters = nb_clusters;
  params.nb_nodes = nb_nodes;

  if ((WINDOW_SIZE*WINDOW_SIZE-1!=X_len) || (3!=Y_len)) 
    {
      printf("input data and filter file not compatible\a\n\n");
      exit(1);
    }

  /*printf("number of leaves = %d\n",nb_clusters);
    printf("number of nodes    = %d\n",nb_nodes);*/
 
 
  for (ctr_i=0;ctr_i<N_coarse;ctr_i++)
    for (ctr_j=0;ctr_j<M_coarse;ctr_j++)
      {
	get_coarse_index(ci_index, cj_index, ctr_i, ctr_j, N_coarse, M_coarse);
	get_fine_index(fi_index,fj_index, ctr_i, ctr_j);
	
	/* center pixel value to be subtracted and added
	   back on to predicted y 
	*/
	ctr_pixel_val = coarse[ci_index[X_len/2]][cj_index[X_len/2]];
	/* just check */
	if ((ci_index[X_len/2] != ctr_i) || (cj_index[X_len/2] != ctr_j))
	  {
	    fprintf(stderr,"ci_index[X_len/2] != ctr_i or cj_index[X_len/2] != ctr_j\n\a\n");
	    exit(-1);
	  }

	/* x, causal part of coarse scale window */
	for (k=0;k<X_len/2;k++)
	  X_data[k]=coarse[ci_index[k]][cj_index[k]];
	/* x, anticausal part, skip center pixel*/ 
	for (k=X_len/2+1;k<=X_len;k++)
	  X_data[k-1]=coarse[ci_index[k]][cj_index[k]];

	/* subtract center pixel values */
	for (k=0;k<X_len;k++)
	  X_data[k] -= ctr_pixel_val;

	
	for (k=0;k<Y_len;k++)
	  Y_data[k]=0.0;
	
	cl=classify_xvec(X_data, X_len, clust, &params);


	/* fixed variance per cluster */
	for (i=0;i<Y_len;i++)
	  for (j=0;j<Y_len;j++)
	    Y_var[i][j] = clust[cl].yer2[i][j];

	
	/* Perform prediction y_hat = a*x     
	   Note that x = X_data is normalized by cr_mean,
	   so result y_hat will also be normalized */
        for (j=0;j<Y_len;j++)
	  for (k=0;k<X_len;k++)
	    Y_data[j] += (X_data[k] * clust[cl].a[k][j]);

	
	/* add b, i.e. y_hat = a*x + b    */
	for (j=0;j<Y_len;j++)
	  Y_data[j] += clust[cl].b[j];
	
	cluster_map[ctr_i][ctr_j] = cl;


	/* Un-normalize predicted variance and assign to Zvar */
	for (i=0;i<Y_len;i++)
	  for (j=0;j<Y_len;j++)
	    Zvar[ctr_i][ctr_j][i][j]=Y_var[i][j]*cr_mean*cr_mean;


	/* Un-normalize predicted mean and assign to Z*/
      	for (k=0;k<Y_len;k++)
	  Z[k][ctr_i][ctr_j]=Y_data[k]*cr_mean;

	  
      }

  if (NORM_FLAG==1)
    multifree((void*)coarse,2);

  
  free((void*)X_data);
  free((void*)ci_index);
  free((void*)cj_index);




  return(cr_mean);

}







