#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "../../AGTREE/agtree.h"
#include "../include/interface_agtree.h"
#include "../include/wav_recon.h"
#include "../include/wav_icm.h"
#include "../include/tomo_defs.h"
#include "../include/wav_transform.h"
#include "../include/utils.h"
#include "../ansi_lib_loc/io_mat_dbl.h"
#include "../ansi_lib_loc/allocate.h"
#include "../ansi_lib_loc/nrutil.h"
#include "../ansi_lib_loc/pseudoinvert.h"
#include "../ansi_lib_loc/svdcmp.h"
#include "../ansi_lib_loc/assort.h"
#include "../ansi_lib_loc/solve2.h"
#include "../ansi_lib_loc/A_comp_dbl_storeflt.h"


#ifdef __linux__
#include <getopt.h>
#endif
extern int optind;
extern char *optarg;


/*#define WRITE_INTERMED */ 




 /* reconstruction routines */
void calc_recon_geometry(struct geom_info *geom,struct all_data *data, struct cmdline_params *cmdline);
void calc_prior_fixvar(double ***fr_prior_params,struct geom_info *geom,int k,double **Xcoarse, struct cluster_info *cluster_info, struct cfg_info *cfg_info, int **mask_coarse);
double calc_true_llikelihood(struct all_data *data,double *S);
void ML_recon(struct all_data *data,double *S,struct geom_info *geom, double **profiles, struct cfg_info *cfg_info,struct column *sp_col,int recon_lev);
void recon(struct all_data *data,struct geom_info *geom,struct column *sp_col, struct cluster_info *cluster_info, struct cfg_info *cfg_info);
void updateZ(struct all_data *data,double **Z_dir,int **Z_coord,double diff, int recon_lev);
void wav_ICD(struct all_data *data,double *S,struct geom_info *geom, double **profiles, struct cfg_info *cfg_info,struct column *sp_col,int lev,struct cluster_info *cluster_info);



/* routines to get input parameters */
void cmdline_error();
void parse_cmdline(struct cmdline_params *cmdline,int argc,char **argv);


/* initialization routines */
double calc_init_const(struct all_data *data,struct geom_info *geom,double **profiles,struct column *sp_col, int recon_lev);
void init_cluster_map(int ***cluster_map,struct geom_info *geom);
void init_pyramid_zero(double ***X, double ****Z,struct geom_info *geom);
void wav_init_sp_col(struct column *sp_col,struct sino_info *sino);


/* memory allocation and deallocation */
void allocate_mscal(struct geom_info *geom,struct all_data *data, struct cluster_info *cluster_info, struct cmdline_params *cmdline, struct ground_truth *ground_truth);
void free_memory(struct all_data *data, struct column *sp_col,struct ground_truth *ground_truth, struct cluster_info *cluster_info, struct cfg_info *cfg_info);



/* data I/O routines */
void load_mask_groundtr(struct all_data *data,struct cmdline_params *cmdline,struct ground_truth *ground_truth, struct geom_info *geom);
void mr_write_output( struct cmdline_params *cmdline,struct all_data *data, struct geom_info *geom, struct ground_truth *ground_truth, struct cluster_info *cluster_info);
void read_cluster_info(struct cluster_info *cluster_info);
double** read_ground_truth(char *fname, int *N, int *M);
void read_sino_data(struct all_data *data,struct column *sp_col,struct cmdline_params *cmdline);






void main( int argc, char **argv ) 
{
  struct geom_info geom[PYR_LEVELS];
  struct column sp_col; 
  struct cmdline_params cmdline;
  struct all_data data;
  struct cfg_info cfg_info;
  struct cluster_info cluster_info;
  struct ground_truth ground_truth;

 
 
  /* parse command-line */
  parse_cmdline(&cmdline,argc,argv);
  
  /* set variables in config struct to values of cmdline */
  cfg_info.no_iter_rec=cmdline.no_iter_rec;
  cfg_info.rel_Ecvar=cmdline.rel_Ecvar;
  cfg_info.w_reg=cmdline.w_reg;
  cfg_info.mask_FLAG=cmdline.mask_FLAG;
  cfg_info.groundtr_FLAG=cmdline.groundtr_FLAG;
  if (cmdline.attn_FLAG)
    sp_col.A_cor.mode='e'; /* attn correction on */
  else
    sp_col.A_cor.mode='t';
  
  
  /* Read sinogram and attenuation data */
  read_sino_data(&data,&sp_col,&cmdline);

  /* Calculate reconstruction geometry for all scales */
  calc_recon_geometry(geom,&data,&cmdline);

  /* Allocate multires pyramids */
  allocate_mscal(geom,&data,&cluster_info,&cmdline,&ground_truth);


  /* read mask and ground truth files,
     perform wavelet decomposition of ground truth 
     and decimation of masks */
  load_mask_groundtr(&data,&cmdline,&ground_truth,geom);

  /* set scatter to minimum if zero */
  if(data.scatter == 0.0)
    data.scatter = 0.001;
  
  
  /* Allocate memory in sp_col*/
  wav_init_sp_col(&sp_col,&data.sino);
  

  /* MAIN tomographic reconstruction */
  recon(&data,geom,&sp_col,&cluster_info,&cfg_info);

  /* write output file */
  mr_write_output(&cmdline,&data,geom,&ground_truth,&cluster_info);
  
  /* free memory */
  free_memory(&data, &sp_col,&ground_truth,&cluster_info,&cfg_info);
  
}






void recon(struct all_data *data,struct geom_info *geom,struct column *sp_col, struct cluster_info *cluster_info, struct cfg_info *cfg_info)
{
  
  int k,i,j,recon_lev;
  int nb_iter_base;
  double **profiles;
  static double *S;
  double init_const;

#ifdef WRITE_INTERMED
  char fname[255];
#endif

  /* save specified number of base iterations */
  nb_iter_base = cfg_info->no_iter_rec;
 
  /* Read prediction tree and nonlinear predictors
     from tree_dat subdirectory */
  init_cluster_map(cluster_info->cluster_map,geom);
  read_cluster_info(cluster_info);
  
 
  /* Allocate state vector S which will contain forward
     projection Ax at any time */
  S = (double *) get_spc(data->sino.n_t*data->sino.n_theta, sizeof(double));


  
  /* Pre-compute projection matrices for all scales, for that we
     first need to compute pixel profiles at scale zero.
     Note, that the proj. matrices for the coarser scales
     are computed by decimating the matrix for the previous
     finer scale. So must pre-compute matrices in order from fine to 
     coarse scale */
  printf("Computing pixel profiles...\n");fflush(stdout);
  pix_prof_comp(&data->sino,&geom[0],&profiles);
  for (k=0;k<=PYR_LEVELS-1;k++)
    A_comp(0,0,&data->sino,&geom[k],profiles,sp_col,k);
  
  /* Start with coarsest scale reconstruction */
  recon_lev=PYR_LEVELS-1;
  

  /* First, initialize all wavelet coeffs Z and X to zero */
  init_pyramid_zero(data->Xall,data->Z,geom);


  /* Now initialize scaling coeffs at coarsest scale 
     with constant image that matches total counts */
  init_const=calc_init_const(data,geom,profiles,sp_col,recon_lev);
  printf("Init coarsest scaling coeffs to const=%f\n\n",init_const);

  for (i=0;i<geom[recon_lev].n_y;i++)
    for (j=0;j<geom[recon_lev].n_x;j++)
      data->Xall[recon_lev][i][j] = init_const;
  
  /* forward project initial condition */
  printf("Calculating forward projection...");fflush(stdout);
  f_pro(S,data,&geom[recon_lev],profiles,sp_col,recon_lev); 
  printf("done\n\n");
     
  
  /* set number of iteration for coarsest scale recon */
  cfg_info->no_iter_rec = (int)(nb_iter_base*pow(1.4142,(double)recon_lev)+0.5);

  /* ML reconstruction of scaling coeffs at coarsest scale */
  printf("ML reconstruction at coarsest scale (%d x %d pixels):\n",geom[recon_lev].n_y,geom[recon_lev].n_x);
  ML_recon(data,S,geom, profiles, cfg_info,sp_col,recon_lev);

  /* Free A-matrix of recon_lev */      
  A_comp(-1,-1,&data->sino,&geom[recon_lev],profiles,sp_col,recon_lev);
 


  /* MAP reconstructions at finer scales, 
     coarse-to-fine loop over scales */
  for (recon_lev=PYR_LEVELS-2;recon_lev>=0;recon_lev--)
    {
#ifdef WRITE_INTERMED
      sprintf(fname,"mv intermed.mat intermed%d.mat",recon_lev);
      system(fname);
#endif
  
      /* set number of iterations for scale */
      if (recon_lev>0)
      cfg_info->no_iter_rec = (int)(nb_iter_base*pow(1.4142,(double)recon_lev)+0.5);
      else
	cfg_info->no_iter_rec = nb_iter_base;


      printf("\nReconstructing at scale %d (%d x %d pixels)\n",recon_lev,geom[recon_lev].n_y,geom[recon_lev].n_x);
      
      /* copy DC coeffs from last coarser reconstructions
	 into reconstruction scale. The wavelet coeffs. at the
	 reconstruction scale are still zero from init_pyramid_zero,
	 so everything is consistent */
      wav_copy_DC(data->Xall,recon_lev,geom);

      /* calculate forward projection at reconstruction scale */
      f_pro(S,data,&geom[recon_lev],profiles,sp_col,recon_lev);
  
      /* Reconstruction */
      wav_ICD(data,S,geom,profiles, cfg_info , sp_col, recon_lev, cluster_info);
      
      /* Free A-matrix of recon_lev */      
      A_comp(-1,-1,&data->sino,&geom[recon_lev],profiles,sp_col,recon_lev);
    }
  
  
  fflush(stdout);fflush(stderr);
  free((void *)S);
  free_img((void **)profiles);
}




/* This is the main reconstruction routine for each scale */
void wav_ICD(struct all_data *data,double *S,struct geom_info *geom, double **profiles, struct cfg_info *cfg_info,struct column *sp_col,int recon_lev,struct cluster_info *cluster_info)
{
  int N,M,iter,s,i,j,k;
  int FLAG;
  struct wavICMdata wICMdata;
  double **Z_dir;
  int **Z_coord;
  double diff;
  double oldx,alpha_new;
  int newt_iter,nb_newt_iter=1;
  MATFile *fout;
  double **X;
  
  double old_l_like,l_like, diff_llike;


  N=geom[recon_lev].n_y;
  M=geom[recon_lev].n_x; 
  X=data->Xall[recon_lev];
  fout=NULL;


  /* At each ICD pixel update, Z_coord[level][k] contains the 
     coordinate of the wavelet tree node at scale level 
     which is affected by the pixel that is updated. 
     So when the pixel is perturbed, only the three
     wavelet coeffs in location
     (Z_coord[level][0],Z_coord[level][1]) are affected
     at scale=level. This implementation 
     is specific to the Haar basis.*/
  Z_coord=(int**)multialloc(sizeof(int),2,PYR_LEVELS-1,2); /*[lev][0:i, 1:j]*/

  /* Z_dir[level][w] contains the partial derivative of the 
     wavelet coefficient w at position Z_coord[level][]
     with respect to the pixel that is being updated */
  Z_dir=(double**)multialloc(sizeof(double),2,PYR_LEVELS-1,3);

 
  /* Re-adapt wavelet graph prior at all coarser scales
     and compute wavelet graph prior at reconstruction
     scale for the first time */
  for (k=PYR_LEVELS-2;k>=recon_lev;k--)
    calc_prior_fixvar(data->allprior_params[k],geom,k,data->Xall[k+1], cluster_info, cfg_info, data->mask[k+1]); 
 
  /* Calculate log-posterior */
  l_like=wav_calc_llikelihood(data,geom, S,recon_lev);
  old_l_like=l_like;
  printf("Begin with normalized log-posterior %f\n",l_like/(N*M));
  fflush(stdout);fflush(stderr);
  
  /* CAREFUL: Within this loop, coarse scale DC coeffs are NOT
     updated, so don't use them.
  */
#ifdef WRITE_INTERMED      
  fout = matOpen("intermed.mat", "w");
  if (fout==NULL)
    {
      fprintf(stderr,"Can't open file intermed.mat for writing.\n\a\n");
      exit(-1);
    }	  
  write_mat_double_image(X,N,M,"img",fout);
  matClose(fout);
#endif

  
  for (iter=0;iter<cfg_info->no_iter_rec;iter++)
    {     
      printf("Starting iter %d\n",iter+1); fflush(stdout);fflush(stderr);
      
      for (s=0;s<N*M;s++) /* loop over pixel locations */
	{
	  /* determine i,j image location, depending on iter go in*/
	  /* row or column order and also start from different corners */
	  imgcoord(&i,&j,s,iter,N,M);
	  
	  /* alpha is the difference between the updated pixel
	     value and the old pixel value */
	  wICMdata.alpha=0.0;
	  oldx=X[i][j];

	  /* compute column of projection matrix for loc. i,j */
	  A_comp(i,j,&data->sino,&geom[recon_lev],profiles,sp_col,recon_lev);
	  
	  /* if more than 0 projections intersect location i,j */
	  if(sp_col->n_row)
	    {
	      /* Calculate Z_dir and Z_coord */
	      calc_Z_dir(Z_dir,Z_coord,i,j,recon_lev);

	      /* Now calculate the first and second partial derivatives
		 of the wavelet prior with respect to the pixel that
		 is being updated. Here the wavelet prior is the Gaussian
		 prior with fixed adaptation. The partial derivatives
		 are stored in wICMdata->pri1 and wICMdata->pri2 .*/
	      calc_prior_params_ls(&wICMdata, Z_dir,data,Z_coord,recon_lev,cluster_info, geom);
	      
	      /* Now we have the prior down to a second order
		 ploynomial in the space domain as a function
		 of the currently updated pixel + const.
		 Thus, we can perform standard ICD optimization
		 in the space domain on the scaling coeffs
		 at the reconstruction scale. 
		 nb_newt_iter says how many Gauss-Seidel 
		 iterations to perform
		 for the pixel and is typically set to 1
	      */
	      for (newt_iter=0;newt_iter<nb_newt_iter;newt_iter++)
		{
		  /* calculate update for pixel value where
		     alpha_new is the difference between the new
		     and the old pixel value */
		  alpha_new=wav_ICMstep(&wICMdata,data,S,sp_col,&FLAG);
		   
		  /* If the new pixel value would be <0, enforce
		     positivity and clip alpha such that new pix-value=0 */
		  if (alpha_new+oldx < 0.0)
		    alpha_new = -1.0*oldx;
		  
		  /* update pixel value */
		  X[i][j] = oldx + alpha_new;  

		  
		  /* FLAG != 0 indicates that BEFORE this iteration,
		     there was a a zero forward projection S[i] for 
		     which Y[i]>0, i.e. log-like derivatives couldn't be
		     calculated correctly. If now X[i][j]>0.0, then
		     this is fixed already, i.e. new S[i]>0, however
		     if X[i][j]==0.0, then need to replace it by positive
		     value to make S[i]>0
		  */
		  if ((FLAG) && (X[i][j] < 1.0e-20))
		    {
		      X[i][j] = 1.0/sp_col->n_row;
		      alpha_new += 1.0/sp_col->n_row;
		    }
		     
		  /* Calculate difference in pixel value between
		     this GS-iteration to the last GS iteration. */
		  diff=alpha_new - wICMdata.alpha;
		  wICMdata.alpha=alpha_new;

		  /* Update forward projection and wavelet decomposition*/
		  if(diff)
		    {
		      updateS(S,sp_col,diff);
		      updateZ(data,Z_dir,Z_coord,diff,recon_lev);
		      /* Update prior means and coarse scale scaling
			 coeffs */
		      update_prior_means(data, diff, recon_lev, cluster_info, Z_coord,geom);
      
		    }
		  else
		    break;
		  
		} /* end for newt-iter */
	    } /* end if sp_col */
	} /* end for s */
      
      /* Calculate log-posterior after full ICD iter */
      l_like=wav_calc_llikelihood(data,geom, S,recon_lev);
      printf("Normalized log-posterior %f\n",l_like/(N*M)); 
      fflush(stdout);fflush(stderr);
      diff_llike=(l_like-old_l_like)/(double)(N*M);
      
      /* for debugging */
      /*if (diff_llike < -1.0e-4)
	{
	fprintf(stderr,"Error: New log-posterior < old log posterior.\n");
	exit(-1);
	} 
      */

#ifdef WRITE_INTERMED     
      if (5*((iter+1)/5)==iter+1)
	{
	  fout = matOpen("intermed.mat", "w");
	  if (fout==NULL)
	    {
	      fprintf(stderr,"Can't open file intermed.mat for writing.\n\a\n");
	      exit(-1);
	    }
	  write_mat_double_image(X,N,M,"img",fout);
	  matClose(fout);
	}
#endif

    } /* end for iter */


  /* Now update coarser scale DC coeffs 
     and recompute wavelet decomposition to avoid accumulation of
     rounding errors*/
  wavelet_decomposition(data->Z,data->Xall,data->Xall[recon_lev],geom,recon_lev);
  
  multifree((void*)Z_dir,2);
  multifree((void*)Z_coord,2);
  
}

 





/* means not log-posterior but only llike term */
double calc_true_llikelihood(struct all_data *data,double *S)
{
  double l_like = 0.0;
  int i;
 
    
  for(i=0;i<data->sino.n_t*data->sino.n_theta;i++)
    if ( S[i] > 0.0 )
      l_like += (-S[i] + data->Y[0][i]*log( S[i] ));
    else 
      if (data->Y[0][i]>1.0e-10)
	l_like -= 1.0e6;

  return(l_like);
}





/* ML_recon is similar to wav_ICD, just that no prior is used,
   so ICD is performed using only likelihood term. This routine
   is used for coarsest scale reconstruction only */
void ML_recon(struct all_data *data,double *S,struct geom_info *geom, double **profiles, struct cfg_info *cfg_info,struct column *sp_col,int recon_lev)
{
  int N,M,iter,s,i,j;
  int FLAG;
  struct wavICMdata wICMdata;
  double diff;
  double oldx;
  MATFile *fout;
  double **X;
  double old_l_like,l_like, diff_llike;

 
  N=geom[recon_lev].n_y;
  M=geom[recon_lev].n_x; 
  X=data->Xall[recon_lev];
  fout=NULL;

  l_like=calc_true_llikelihood(data, S);
  old_l_like=l_like;
  printf("Begin with normalized log-likelihood %f\n",l_like/(N*M));
  fflush(stdout);fflush(stderr);



  /* CAREFUL: Within this loop, coarse scale DC coeffs are NOT
     updated, so don't use them.
  */
  for (iter=0;iter<cfg_info->no_iter_rec;iter++)
    {     
      printf("Starting iter %d\n",iter+1); fflush(stdout);fflush(stderr);
      
      for (s=0;s<N*M;s++) /* loop over pixel locations */
	{
	  /* determine i,j image location, depending on iter go in*/
	  /* row or column order and also start from different corners */
	  imgcoord(&i,&j,s,iter,N,M);

	  wICMdata.alpha=0.0;
	  oldx=X[i][j];

	  /* compute column of A matrix for loc. i,j */
	  A_comp(i,j,&data->sino,&geom[recon_lev],profiles,sp_col,recon_lev);
	  
	  /* if more than 0 projections intersect location i,j */
	  if(sp_col->n_row)
	    {
	      FLAG = wav_compute_derivatives(&wICMdata, data, sp_col,S);
	      X[i][j] = oldx - wICMdata.D1/wICMdata.D2;
	      
	      if (X[i][j]<0.0)
		X[i][j]=0.0;

	      if ((FLAG) && (X[i][j] < 1.0e-20))
		{
		  X[i][j] = 1.0/sp_col->n_row;
		}
	      
	      
	      diff=X[i][j]- oldx;
	      if(diff)
		{
		  updateS(S,sp_col,diff);
		}
	      
	    } /* end if sp_col */
	} /* end for s */
      
     
      /* check log-posterior */


      l_like=calc_true_llikelihood(data,S);
      printf("Normalized log-likelihood %f\n",l_like/(N*M)); 
      fflush(stdout);fflush(stderr);
      diff_llike=(l_like-old_l_like)/(double)(N*M);
      
      /* For debugging */
      /*if (diff_llike < -1.0e-4)
	{ 
	fprintf(stderr,"Error: New log-posterior < old log posterior.\n");
	exit(-1);
	}
      */
      
#ifdef WRITE_INTERMED
      if (5*((iter+1)/5)==iter+1)
	{
	  fout = matOpen("intermed.mat", "w");
	  if (fout==NULL)
	    {
	      fprintf(stderr,"Can't open file intermed.mat for writing.\n\a\n");
	      exit(-1);
	    }

	  write_mat_double_image(X,N,M,"img",fout);
	  matClose(fout); 
	} 
#endif

    } /* end for iter */
}

 
    


/* Perform adaptation step of adaptive wavelet graph model
   for a single scale k. The prior for level k is determined
   by first classifying the scaling coeffs at scale k+1 using the
   agtree classifier. Then, based on the classification, we calculate
   the prior means (using scaling coeffs at scale k+1) and the 
   prior covariances for the wavelet coeffs at scale k.
   Note, that the prior means are not fixed, but will 
   change during the wav_ICD reconstruction.
   k is fine scale level */
void calc_prior_fixvar(double ***fr_prior_params,struct geom_info *geom,int k,double **Xcoarse, struct cluster_info *cluster_info, struct cfg_info *cfg_info, int **mask_coarse)
{ 
  int i,j,N,M,w,n,m;

  double ***tmp_img;
  double ****tmp_img2;
  double **Rinv;


  static struct node **clust;
  static int *nb_clusters,*nb_nodes;
  int **cluster_map;
  double cr_mean;
  double **scal_Ecvar;
  double sig2;
  double rel_Ecvar;

  clust=cluster_info->clust;
  nb_clusters=cluster_info->nb_clusters;
  nb_nodes=cluster_info->nb_nodes;
  cluster_map=cluster_info->cluster_map[k];
  

  N=geom[k].n_y; /* fine scale params */
  M=geom[k].n_x;

  tmp_img=(double***)multialloc(sizeof(double),3,3,N/2,M/2);/* mean */
  tmp_img2=(double****)multialloc(sizeof(double),4,N/2,M/2,3,3); /* Covar */
  Rinv=(double**)multialloc(sizeof(double),2,3,3);
  scal_Ecvar=(double**)multialloc(sizeof(double),2,3,3);

 
  /* perform classification, get back means (tmp_img) and 
     covariance matrices (tmp_img2). The returned values
     tmp_img and tmp_img2 are already un-normalized, i.e. 
     are at the correct amplitude level of the reconstruction.
     However, the expected conditional wavelet-coeff. covariance 
     Ecvar have been read in earlier and apply to normalized
     wavelet coeffs with a trimmed mean of 1. 
     Thus we also return the normalization constant 
     cr_mean to scale up the exp. conditional covariance matrix */
  cr_mean=wav_predict_fixvar(tmp_img, tmp_img2, Xcoarse, geom[k+1].n_y, geom[k+1].n_x,clust[k],nb_clusters[k],nb_nodes[k],cluster_map,k+1,mask_coarse,cfg_info->mask_FLAG);
  
  
  /* copy means */
  for (i=0;i<N/2;i++)
    for (j=0;j<M/2;j++)
      for (w=0;w<3;w++)
	fr_prior_params[w][i][j] = tmp_img[w][i][j];


  /* Compute covariance matrices for the triplet of 
     wavelet coeffs at each location in scale k*/
  for (i=0;i<N/2;i++)
    for (j=0;j<M/2;j++)
      {
	/* Scale the uncoditional variances of the 3 wavelet-coeffs
	   by the trimmed_mean^2 */
	for (n=0;n<3;n++)
	  for (m=0;m<3;m++)
	    scal_Ecvar[n][m] = cr_mean * cr_mean * cluster_info->stat_var[k][n][m];
	/* sig2 is overall variance scaling = regularization parameter */
	sig2 = cfg_info->w_reg * cfg_info->w_reg;

	/* rel_Ecvar is between 0 and 1 and determines the 
	   relative weighting of the exp. conditional variance
	   as compared to the cluster variances */
	rel_Ecvar = cfg_info->rel_Ecvar;

	/* compute non-inverted covariance matrix */
	for (n=0;n<3;n++)
	  for (m=0;m<3;m++)
	    Rinv[n][m] = sig2 * ( (1.0-rel_Ecvar)*tmp_img2[i][j][n][m] + rel_Ecvar*scal_Ecvar[n][m] );
	

	/* invert */
	pseudoinvert(Rinv,3);
	
	/* Copy into prior_params struct */
	for (n=0;n<3;n++)
	  for (m=0;m<3;m++)
	    fr_prior_params[3*n+m+3][i][j] = Rinv[n][m];
      }


  multifree((void*)Rinv,2);
  multifree((void*)tmp_img,3);
  multifree((void*)tmp_img2,4);
  multifree((void*)scal_Ecvar,2);
}  




double** read_ground_truth(char *fname, int *N, int *M)
{
  MATFile *fin;  
  double **X_true;
  
  /* read input image  */
  fin = matOpen(fname, "r");
  if (fin == NULL)
    {
      fprintf(stderr,"Cannot read MatLab file %s\n",fname);
      exit(-1);
    }
  
  X_true = read_mat_double_image(N,M,"img",fin);
  if (X_true == NULL)
    {
      fprintf(stderr,"Cannot read img matrix in infile %s \n",fname);
      exit(-1);
    }
  matClose(fin);

  return(X_true);

}

 

void mr_write_output(struct cmdline_params *cmdline, struct all_data *data,struct geom_info *geom, struct ground_truth *ground_truth, struct cluster_info *cluster_info)
{
  MATFile *fout;
  int k,N,M,w;
  char matrix[40];
  double ***Xtmp;
  char *fname;
  int ***cluster_map;
  double ****prior_params;
  double ***X;
  double ****Z;

  fname=cmdline->output;
  cluster_map=cluster_info->cluster_map;
  prior_params=data->allprior_params;
  
  X=data->Xall;
  Z=data->Z;

  printf("Writing output image: %s\n",fname);
  
  
  fout = matOpen(fname, "w");
  if (fout == NULL)
    {
      fprintf(stderr,"Cannot write MatLab file %s\n",fname);
      exit(-1);
    }

  for (k=0;k<PYR_LEVELS;k++)
    {
      N=geom[k].n_y;
      M=geom[k].n_x;

      /* write reconstruction */
      sprintf(matrix,"X%d",k);
      write_mat_double_image(X[k],N,M,matrix,fout);
      
      /* write true scaling coeffs */
      if (cmdline->groundtr_FLAG)
	{
	  sprintf(matrix,"Xt%d",k);
	  write_mat_double_image(ground_truth->X_true[k],N,M,matrix,fout);
	}    
    }

  if (cmdline->diagoutput_FLAG)
    {  
      Xtmp=(double***)mget_spc(PYR_LEVELS,sizeof(double**));
      for (k=0;k<PYR_LEVELS;k++)
	Xtmp[k]=(double**)multialloc(sizeof(double),2,geom[k].n_y,geom[k].n_x);
      
      for (k=0;k<PYR_LEVELS-1;k++)
	{
	  for (w=0;w<3;w++)
	    {
	      N=geom[k].n_y;
	      M=geom[k].n_x;
	      
	      /* write reconstructed wavelet coeffs */
	      sprintf(matrix,"Z%d_%d",k,w);
	      write_mat_double_image(Z[k][w],N/2,M/2,matrix,fout);
	      
	      /* write predicted means of wavelet coeffs */
	      sprintf(matrix,"Zm%d_%d",k,w);
	      write_mat_double_image(prior_params[k][w],N/2,M/2,matrix,fout);
	      
	      /* write inverse variance */
	      sprintf(matrix,"Zivar%d_%d",k,w);
	      write_mat_double_image(prior_params[k][4*w+3],N/2,M/2,matrix,fout);
	      
	      /* write ground truth wavelet coeffs */
	      if (cmdline->groundtr_FLAG)
		{
		  sprintf(matrix,"Zt%d_%d",k,w);
		  write_mat_double_image(ground_truth->Z_true[k][w],N/2,M/2,matrix,fout);
		}
	    }
	  
	  
	}
 

      /* Now write the mean prediction, compute by 
	 adding the predicted means of the wavelet coeffs
	 in prior_params to the scaling coeffs in X at the next coarser
	 scale */
      wavelet_reconstruction_trdc(Xtmp,prior_params,X,geom);
      for (k=0;k<PYR_LEVELS-1;k++)
	{
	  N=geom[k].n_y;
	  M=geom[k].n_x;
	  sprintf(matrix,"Xip%d",k);
	  write_mat_double_image(Xtmp[k],N,M,matrix,fout);
	}
      
      for (k=0;k<PYR_LEVELS;k++)
	multifree((void*)Xtmp[k],2);
      free((void*)Xtmp);
    }

  matClose(fout);
}
 






void cmdline_error()
{
  fprintf(stderr,"Usage:\n\a\n");
  fprintf(stderr,"tomo [-i no_iter at finest scale (15)] [-r resolution (256)] [-f field_of_view_mm (n_t*delta_t)] [-s regularization parameter (1.0)] [-w relative weight of expected conditional covar (0.001)] [-T attn_file] [-t gound_truth_file] [-d write_wav_decomp] input.mat mask.mat output.mat\n\n");
  exit(-1);   
}


void parse_cmdline(struct cmdline_params *cmdline,int argc,char **argv)
{
  int ct=0;
  int i,rs;

  cmdline->no_iter_rec=15;
  cmdline->w_reg=1.0;
  cmdline->rel_Ecvar=0.001;
  cmdline->resol=256;
  cmdline->fov=1.0;

  cmdline->mask_FLAG=1; /* no longer an option, always use mask*/
  cmdline->groundtr_FLAG=0;
  cmdline->attn_FLAG=0;
  cmdline->fov_FLAG=0;
  cmdline->diagoutput_FLAG=0;

  for (;;)
    { 
      i= getopt(argc, argv, "i:s:w:t:r:f:T:d");
      if (i==EOF)
        break;
      else
        {
          if (i == '?')
            cmdline_error();
	  if (i=='i')
	    cmdline->no_iter_rec=atoi(optarg);
	    
	  if (i=='s')
	    cmdline->w_reg=atof(optarg);
	  
	  if (i=='w')
	    cmdline->rel_Ecvar=atof(optarg);
	  
	  if (i=='t')
	    {
	      cmdline->groundtr_FLAG=1;
	      strcpy(cmdline->ground_truth,optarg);
	    }
  
	  if (i=='r')
	    cmdline->resol=atoi(optarg);
	    
	  if (i=='f')
	    {
	      cmdline->fov=atof(optarg);
	      cmdline->fov_FLAG=1;
	    }
	  
	  if (i=='T')
	    {
	      cmdline->attn_FLAG=1;
	      strcpy(cmdline->attenuation,optarg);
	    }
	  if (i=='d')
	    cmdline->diagoutput_FLAG=1;
	  

	}
    }

  ct=optind;
  if ((argc-optind < 3) || (argc-optind > 3))
    {
      fprintf(stderr,"\nError: Wrong number of arguments.\n\a\n");
      cmdline_error();
    }
 
  cmdline->input   = argv[ct];
  ct++;
  
  cmdline->mask_file = argv[ct];
  ct++;

  cmdline->output = argv[ct];
  ct++;


  /* check input parameters */
  /* is resolution power of two or contains enough factors of
     2 for PYR_LEVELS ? */
  rs=cmdline->resol;
  for (i=1;i<PYR_LEVELS;i++)
    rs=rs/2;
  for (i=1;i<PYR_LEVELS;i++)
    rs=rs*2;
  if (rs!=cmdline->resol)
    {
      fprintf(stderr,"Error: resolution of reconstruction must be power of two or contain factor of 2^(PYR_LEVELS-1)=%d\n",(int)(pow(PYR_LEVELS-1,2)+0.5));
      cmdline_error();
    }

  /* at least one iteration? */
  if (cmdline->no_iter_rec<1)
    {
      fprintf(stderr,"Error: specified number of iterations is %d, must be greater or equal to one.\n",cmdline->no_iter_rec);
      cmdline_error();
    }
  
  
  /* overall reg. parameter must be >0 */
  if (cmdline->w_reg<1e-10)
    {
      fprintf(stderr,"Error: regularization parameter (-s) must be greater than 1e-10. Specified value is %g.\n",cmdline->w_reg);
      cmdline_error();
    } 

  /* relative weight of exp. conditional variance must be between 0 and one */
  if ((cmdline->rel_Ecvar<0.0) || (cmdline->rel_Ecvar>1.0))
    {
      fprintf(stderr,"Error: relative weight (-w) of exp. conditional covariance must be between zero and one. Specified value is %g.\n",cmdline->rel_Ecvar);
      cmdline_error();
    } 

  if ((cmdline->fov_FLAG) && (cmdline->fov <=0.0))
    {
      fprintf(stderr,"Error: field of view must be >0. Specified value is %g.\n",cmdline->fov);
      cmdline_error();
    } 



  printf("\nInput parameters:\n");
  printf("  sinogram file: %s\n",cmdline->input);
  printf("  output file  : %s\n",cmdline->output);
  printf("  reconstruction resolution: %d\n",cmdline->resol);
  printf("  number of iterations (finest scale) : %d\n",cmdline->no_iter_rec);
  printf("  overall regularization parameter: %f\n",cmdline->w_reg);
  printf("  relative weight of exp. conditional covariance: %f\n",cmdline->rel_Ecvar);
  if (cmdline->mask_FLAG)
    printf("  mask file : %s\n",cmdline->mask_file);
  else
    printf("  NO MASK SPECIFIED (not recommended)\n");

  if (cmdline->fov_FLAG)
    printf("  field of view (mm) : %f\n",cmdline->fov);
  else
    printf("  default field of view (n_t*delta_t)\n");
  
  if (cmdline->groundtr_FLAG)
    printf("  ground truth file : %s\n",cmdline->ground_truth);
  
  if (cmdline->attn_FLAG)
    printf("  attenuation correction file : %s\n",cmdline->attenuation);
  else
    printf("  not using attenuation correction.\n");
}
 










 
void wav_init_sp_col(struct column *sp_col,struct sino_info *sino)
{
  sp_col->val = (double *) get_spc(sino->n_t*sino->n_theta, sizeof(double));
  sp_col->row = (int *) get_spc(sino->n_t*sino->n_theta, sizeof(int));
  sp_col->multiplier = 1.0;
}





void free_memory(struct all_data *data, struct column *sp_col,struct ground_truth *ground_truth, struct cluster_info *cluster_info, struct cfg_info *cfg_info)
{
  int k;
  double ***X;
  double ****Z;
  int ***cluster_map;
  double ****prior_params;


  X=data->Xall;
  Z=data->Z;
  cluster_map=cluster_info->cluster_map;
  prior_params=data->allprior_params;

  printf("Freeing memory ...\n");

  for (k=0;k<PYR_LEVELS;k++)
    free_img((void **)X[k]);
  free((void*)X);
  
  
  for (k=0;k<PYR_LEVELS-1;k++)
    {
      multifree((void*)Z[k],3);
      multifree((void*)prior_params[k],3);
      free_img((void **)cluster_map[k]);
    }
  
  free((void*)Z);
  free((void*)prior_params);
  free((void*)cluster_map);
  free_img((void **)data->Y);
  free((void *)sp_col->val);
  free((void *)sp_col->row);

  if (cfg_info->groundtr_FLAG)
    {
      for (k=0;k<PYR_LEVELS;k++)
	free_img((void **)ground_truth->X_true[k]);
      for (k=0;k<PYR_LEVELS-1;k++)
	multifree((void*)ground_truth->Z_true[k],3);
      free((void*)ground_truth->X_true);
      free((void*)ground_truth->Z_true);
    }
  
  if (cfg_info->mask_FLAG)
    {
      for (k=0;k<PYR_LEVELS;k++) 
	multifree((void*)data->mask[k],2);
    }   
  free((void*)data->mask);

  if (sp_col->A_cor.mode=='e')
    free_img((void **)sp_col->A_cor.Y_t);

}


void read_sino_data(struct all_data *data,struct column *sp_col,struct cmdline_params *cmdline)
{
  MATFile *fin;  
  mxArray *pm;
  int Ang,T;
  
  /* read input sinogram Y */
  fin = matOpen(cmdline->input, "r");
  if (fin == NULL)
    {
      fprintf(stderr,"Cannot read MatLab file %s\n",cmdline->input);
      exit(-1);
    }
  
  data->Y = read_mat_double_image(&Ang,&T,"counts",fin);
  if (data->Y == NULL)
    {
      fprintf(stderr,"Cannot read sino matrix \n");
      exit(-1);
    }
  
  
  read_mat_sino_info(fin, &data->sino);
  
  pm = matGetArray(fin,"scatter");
  if(pm != NULL) {
    data->scatter = mxGetScalar(pm);
  }
  mxDestroyArray(pm);

  
  matClose(fin);
  
  

  /* Read attenuation */
  if (sp_col->A_cor.mode == 'e')
    {
      fin = matOpen(cmdline->attenuation, "r");
      if (fin == NULL)
	{
	  fprintf(stderr,"\a Cannot read MatLab file %s\n",cmdline->attenuation);
	  exit(-1);
	}
    
      sp_col->A_cor.Y_t = read_mat_double_image(&Ang,&T,"trn",fin);
      if (sp_col->A_cor.Y_t == NULL)
	{
	  fprintf(stderr,"Cannot read trn matrix in file %s.\n\a\n",cmdline->attenuation);
	  exit(-1);
	}
      matClose(fin);
      
    }
}
  


void load_mask_groundtr(struct all_data *data,struct cmdline_params *cmdline,struct ground_truth *ground_truth, struct geom_info *geom)
{
  int Ngt,Mgt;
  
  
  /* if mask, read mask */
  if (cmdline->mask_FLAG)
    {
      data->mask[0]=read_mask(&geom[0],cmdline->mask_file);
      decimate_masks(data->mask,&geom[0]);
    }

  
  /* read original image */
  if (cmdline->groundtr_FLAG)
    {
      ground_truth->X_true[0]=read_ground_truth(cmdline->ground_truth, &Ngt,&Mgt);
      /* check size */
      if ((Ngt!=geom[0].n_y) || (Mgt!=geom[0].n_x))
	{
	  fflush(stdout);fflush(stderr);
	  fprintf(stderr,"Error: Size of specified ground truth image is different from reconstruction resolution.\n\a\n");
	  exit(-1);
	}
      
      /* initialize pyramids */
      wavelet_decomposition(ground_truth->Z_true,ground_truth->X_true,ground_truth->X_true[0],geom,0);
    }


  
}
  






double calc_init_const(struct all_data *data,struct geom_info *geom,double **profiles,struct column *sp_col, int recon_lev)
{
  double sumA=0.0,sumY=0.0;
  int i,j,t;
  double init_const;
  
  for (i=0;i<geom[recon_lev].n_y;i++)
    for (j=0;j<geom[recon_lev].n_x;j++)
      {
	A_comp(i,j,&data->sino,&geom[recon_lev],profiles,sp_col,recon_lev);
	for (t=0;t<sp_col->n_row;t++)
	  sumA += sp_col->val[t];
      }
  
  for (i=0;i<data->sino.n_theta;i++)
    for (j=0;j<data->sino.n_t;j++)
      sumY += data->Y[i][j];
  
  /*printf("sumY=%f sumA=%f\n",sumY,sumA);*/
  
  init_const = sumY/sumA;

  printf("Initializing to constant = %g\n\n",init_const);

  return(init_const);
}


void updateZ(struct all_data *data,double **Z_dir,int **Z_coord,double diff, int recon_lev)
{
  int lev,w;

  for (lev=recon_lev;lev<PYR_LEVELS-1;lev++)
    for (w=0;w<3;w++)
      data->Z[lev][w][Z_coord[lev][0]][Z_coord[lev][1]] += (diff*Z_dir[lev][w]);
  

}




void read_cluster_info(struct cluster_info *cluster_info)
{
  char fname[255];
  int i,stat,lev;
  char fname_loc[255];
  FILE *fp;
  double x;
  int n,m;

  sprintf(fname,"tree_dat/clust_l");

  for (i=0;i<PYR_LEVELS-1;i++)
    {
      sprintf(fname_loc,"%s%d.asc",fname,i);
      cluster_info->clust[i]=read_tree(fname_loc, &cluster_info->X_len, &cluster_info->Y_len, &cluster_info->nb_clusters[i], &cluster_info->nb_nodes[i]);
      if ((WINDOW_SIZE*WINDOW_SIZE-1!=cluster_info->X_len) || (3!=cluster_info->Y_len)) 
	{
	  printf("input data and filter file not compatible\a\n\n");
	  exit(1);
	}
    }  




  /* read static variances */
  fflush(stdout);fflush(stderr);
  printf("Reading expected conditional variance info...");fflush(stdout);

  for (lev=0;lev<PYR_LEVELS-1;lev++)
    {
      sprintf(fname_loc,"%s%d.asc_Evar",fname,lev);
      fp = fopen(fname_loc,"r");
      if (fp==NULL)
	{
	  fprintf(stderr,"Error: Can't open file %s for reading.\n\a\n",fname_loc);
	  exit(-1);
	}
      
      for (m=0;m<3;m++)
	for (n=0;n<3;n++)
	  {
	    stat=fscanf(fp,"%lf",&x);
	    if (stat!=1)
	      {
		fprintf(stderr,"Error: Read error in %s. stat=%d. \n\a\n",fname_loc,stat);
		exit(-1);
	      }
	    cluster_info->stat_var[lev][m][n]=x;
	  }
      fclose(fp);

      /*printf("Evar level %d\n",lev);
      for (m=0;m<3;m++)
	{
	  for (n=0;n<3;n++)
	    printf("%g  ",cluster_info->stat_var[lev][m][n]);
	  printf("\n");
	  }*/
    }
  fflush(stdout);fflush(stderr);
  printf("done.\n\n");fflush(stdout);fflush(stderr);
      
}


void init_cluster_map(int ***cluster_map,struct geom_info *geom)
{
  int k,i,j;

  for (k=0;k<PYR_LEVELS-1;k++)
    for (i=0;i<geom[k].n_x/2;i++)
      for (j=0;j<geom[k].n_y/2;j++)
	cluster_map[k][i][j]=-1;


}


void init_pyramid_zero(double ***X, double ****Z,struct geom_info *geom)
{
  int lev,w,i,j;

  for (lev=0;lev<PYR_LEVELS;lev++)
    for (i=0;i<geom[lev].n_y;i++)
      for (j=0;j<geom[lev].n_x;j++)
	X[lev][i][j] = 0.0;
      
  for (lev=0;lev<PYR_LEVELS-1;lev++)
    for (i=0;i<geom[lev].n_y/2;i++)
      for (j=0;j<geom[lev].n_x/2;j++)
	for (w=0;w<3;w++)
	  Z[lev][w][i][j] = 0.0;
    
}

void calc_recon_geometry(struct geom_info *geom,struct all_data *data, struct cmdline_params *cmdline)
{
  int resol,k;

  resol=cmdline->resol;

  if (cmdline->fov_FLAG)
    geom[0].dia=cmdline->fov;
  else
    geom[0].dia = data->sino.n_t*data->sino.delta_t;



  geom[0].n_y = resol;
  geom[0].n_x = resol;
  geom[0].delta = geom[0].dia/(double)resol;
  geom[0].x_0 = (-geom[0].dia+geom[0].delta)/2.0;
  geom[0].y_0 = (-geom[0].dia+geom[0].delta)/2.0;

   for(k=1;k<PYR_LEVELS;k++)
    {
      geom[k].n_y=geom[k-1].n_y/2;
      geom[k].n_x=geom[k-1].n_x/2;
      geom[k].dia = geom[0].dia;
      geom[k].delta = geom[k-1].delta*2.0;
      geom[k].x_0 = (-geom[k].dia+geom[k].delta)/2.0;
      geom[k].y_0 = (-geom[k].dia+geom[k].delta)/2.0;
    }
   




}






void allocate_mscal(struct geom_info *geom,struct all_data *data, struct cluster_info *cluster_info, struct cmdline_params *cmdline, struct ground_truth *ground_truth)
{
  
  int k;

  
  /* X[lev][i][j] is pyramid of scaling coeffs */
  data->Xall = (double***)mget_spc(PYR_LEVELS,sizeof(double**));
  for(k=0;k<PYR_LEVELS;k++)
    data->Xall[k] = (double **)get_img(geom[k].n_x,geom[k].n_y,sizeof(double));

  /* Z[lev][coeff][i][j] is pyramid of wavelet coeffs */
  data->Z  = (double****)mget_spc(PYR_LEVELS-1,sizeof(double***));
  for(k=0;k<PYR_LEVELS-1;k++)
    data->Z[k] = (double ***)multialloc(sizeof(double),3,3,geom[k].n_y/2,geom[k].n_x/2);


  /* prior_params[level][param][row][col]
     param=0..2 : means of wavelet coeffs 1..3
     param=3..11: INVERSE Covariance matrix of wavelet coeffs 1..3
     diag. entries are 3,7,11.       */
  data->allprior_params=(double****)mget_spc(PYR_LEVELS-1,sizeof(double***));
  for(k=0;k<PYR_LEVELS-1;k++)
    data->allprior_params[k] = (double ***)multialloc(sizeof(double),3,12,geom[k].n_y/2,geom[k].n_x/2); 



  /* cluster_map is important for the reconstruction optimization */
  cluster_info->cluster_map = (int***)mget_spc(PYR_LEVELS-1,sizeof(int**));
  for(k=0;k<PYR_LEVELS-1;k++)
    cluster_info->cluster_map[k] = (int **)get_img(geom[k].n_x/2,geom[k].n_y/2,sizeof(int));
  
  /* mask at each scale is used to specify support of 
     reconstructed object. The normalization factors for
     the non-linear prediction are then obtained using
     only pixels in the region of support */
  data->mask = (int***)mget_spc(PYR_LEVELS,sizeof(int**));
  
  
  if (cmdline->groundtr_FLAG)
    {
      ground_truth->X_true = (double***)mget_spc(PYR_LEVELS,sizeof(double**));
      ground_truth->Z_true  = (double****)mget_spc(PYR_LEVELS,sizeof(double***));

      for(k=1;k<PYR_LEVELS;k++)
	ground_truth->X_true[k] = (double **)get_img(geom[k].n_x,geom[k].n_y,sizeof(double));
      for(k=0;k<PYR_LEVELS-1;k++)
	ground_truth->Z_true[k] = (double ***)multialloc(sizeof(double),3,3,geom[k].n_y/2,geom[k].n_x/2);
    }

 


}
