#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "../../AGTREE/agtree.h"

#include "../include/interface_agtree.h"
#include "../include/utils.h"
#include "../include/tomo_defs.h"
#include "../ansi_lib_loc/io_mat_dbl.h"
#include "../ansi_lib_loc/allocate.h"
#include "../include/wav_transform.h"

#ifdef __linux__
#include <getopt.h>
#endif
extern int optind;
extern char *optarg;

struct cmdfile{
  int nb_images;
  char **fnames;
  char **fn_masks;
};



void free_memory(struct cmdfile *cmdfile, double ***X_true, double ****Z_true, double ****prior_params);
void init_pyramids(struct geom_info *geom,double ***X_true, double ****Z_true,double ****prior_params );
void read_cmdfile(int argc,char **argv,struct cmdfile *cmdfile,double *reg);
double** read_image(struct geom_info *geom,char *fname);
void free_mem_img(double ***X_true, double ****Z_true,double ****prior_params,int ***mask);
void train_img(struct geom_info *geom,double ***X_true,double ****Z_true,double ****prior_params, int img, int ***mask);
void cmdline_error();




void main(int argc, char **argv )
{
  struct cmds cmds;
  
  double ***X_true, ****Z_true;
  int ***mask;
  struct geom_info geom[PYR_LEVELS];
  struct cmdfile cmdfile;
  double ****prior_params;
  int img,lev;

  int initial_nb_clusters[PYR_LEVELS-1] = {150,150,150,100};

  double reg; /* regularization factor for training */
  


  /* read_cmdfile and reg-value */
  read_cmdfile(argc,argv,&cmdfile,&reg);
  
  printf("Initial number of clusters (= no. tree leaves) in train.c set to\n");
  for (lev=0;lev<PYR_LEVELS-1;lev++)
    printf("  scale %d: %d\n",lev,initial_nb_clusters[lev]);
  printf("\n");
  
  /* Allocate data pyramids for DC-coeffs X and wavelet coeffs Z */ 
  X_true = (double***)get_spc(PYR_LEVELS,sizeof(double**));
  mask = (int***)get_spc(PYR_LEVELS,sizeof(int**));
  Z_true = (double****)get_spc(PYR_LEVELS-1,sizeof(double***));
  prior_params = (double****)get_spc(PYR_LEVELS-1,sizeof(double***));


  /* loop over training images */
  for (img=0;img<cmdfile.nb_images;img++)
    {
      printf("Extracting training data from image %d\n",img);
      /* read image */
      X_true[0]=read_image(&geom[0],cmdfile.fnames[img]);
      
      init_pyramids(geom,X_true,Z_true,prior_params);
      /* read mask and decimate */
      mask[0]=read_mask(&geom[0],cmdfile.fn_masks[img]);
      decimate_masks(mask,&geom[0]);

      wavelet_decomposition(Z_true,X_true,X_true[0],geom,0);

      /* write training vectors for image to hard disc */
      train_img(geom,X_true,Z_true,prior_params,img,mask);
      free_mem_img(X_true, Z_true,prior_params,mask);
    }



  free_memory(&cmdfile,X_true, Z_true,prior_params);

  for (lev=PYR_LEVELS-1;lev>0;lev--)
    {
      sprintf(cmds.trainfile,"tree_dat/train_dat_l%d.asc",lev-1);
      sprintf(cmds.prunefile,"tree_dat/prune_dat_l%d.asc",lev-1);
      sprintf(cmds.outfile,"tree_dat/clust_l%d.asc",lev-1);
      cmds.prune=1;
      cmds.alpha = reg;
      cmds.merge_limit = 1; 
      cmds.nb_clusters = initial_nb_clusters[lev-1];

      /* make_agtree is a command from libagt.a */
      printf("\nBuilding prediction tree for scale %d:\n",lev-1);
      make_agtree(&cmds);
      

    }

  printf("Training successful. You can now delete the files\ntree_dat/prune_dat_* and tree_dat/train_dat_* .\n\n");

}






void free_mem_img(double ***X_true, double ****Z_true,double ****prior_params, int ***mask)
{
  int lev;
  
  for (lev=0;lev<PYR_LEVELS;lev++)
    {
      free_img((void**)X_true[lev]);
      multifree((void*)mask[lev],2);

    }

  for (lev=0;lev<PYR_LEVELS-1;lev++)
    {
      multifree((void*)Z_true[lev],3);
      multifree((void*)prior_params[lev],3);
    }
  

}

void free_memory(struct cmdfile *cmdfile, double ***X_true, double ****Z_true, double ****prior_params)
{

  
  free((void*)X_true);
  free((void*)Z_true);
  free((void*)prior_params);
  
  multifree((void*)cmdfile->fnames,2);
}








void init_pyramids(struct geom_info *geom,double ***X_true, double ****Z_true,double ****prior_params )
{
  int k;

  for(k=1;k<PYR_LEVELS;k++)
    {
      geom[k].n_y = geom[k-1].n_y/2;
      geom[k].n_x = geom[k-1].n_x/2;
      X_true[k] = (double **)get_img(geom[k].n_x,geom[k].n_y,sizeof(double));
    }

  /* allocate main data pyramids */
  for(k=0;k<PYR_LEVELS-1;k++)
    {
      Z_true[k] = (double ***)multialloc(sizeof(double),3,3,geom[k].n_y/2,geom[k].n_x/2);
      prior_params[k] = (double ***)multialloc(sizeof(double),3,12,geom[k].n_y/2,geom[k].n_x/2); 
    }



}


double** read_image(struct geom_info *geom,char *fname)
{
  MATFile *fin;  
  double **img;
  
  fin = matOpen(fname, "r");
  if (fin == NULL)
    {
      fprintf(stderr,"Cannot read MatLab file %s\n",fname);
      exit(-1);
    }

  img = read_mat_double_image(&geom->n_y,&geom->n_x,"img",fin);
  if (img == NULL)
    {
      fprintf(stderr,"Cannot read img matrix \n");
      exit(-1);
    }
  matClose(fin);
  return(img);
}


void cmdline_error()
{
  fprintf(stderr,"Usage:\n\a\n");
  fprintf(stderr,"train [-r reg-factor (0.011)] cmdfile\n\a\n");
  exit(-1);   
}



void read_cmdfile(int argc,char **argv,struct cmdfile *cmdfile,double *reg)
{
  FILE *fp;
  int i,stat,ct;
  
  *reg=0.011; /* default value */

  for (;;)
    { 
      i= getopt(argc, argv, "r:");
      if (i==EOF)
        break;
      else
        {
          if (i == '?')
            cmdline_error();
	  if (i=='r')
	    *reg = atof(optarg);
	}
    }

  if (*reg<0.0)
    {
      fprintf(stderr,"Error: reg-factor must not be negative. Specified to %f.\n\n",*reg);
      cmdline_error();
    }


  ct=optind;
  if (argc-optind != 1) 
    {
      fprintf(stderr,"Error: wrong number of arguments.\n\n");
      cmdline_error();
    }
  
  fp=fopen(argv[ct],"r");
  if (fp==NULL)
    {
      fprintf(stderr,"Can't open cmdfile %s .\n\a\n",argv[ct]);
      exit(-1);
    }
  
  stat=fscanf(fp,"%d",&cmdfile->nb_images);
  if (stat!=1)
    {
      fprintf(stderr,"Can't read nb_images in cmdfile\n\a\n");
      exit(-1);
    }
  else if (cmdfile->nb_images<1) 
    {
      fprintf(stderr,"Invalid number of images %d in cmdfile\n\a\n",cmdfile->nb_images);
      exit(-1);
    }
  
  cmdfile->fnames = (char**)multialloc(sizeof(char),2,cmdfile->nb_images,256);
  cmdfile->fn_masks = (char**)multialloc(sizeof(char),2,cmdfile->nb_images,256);

  for (i=0;i<cmdfile->nb_images;i++)
    {
      stat=fscanf(fp,"%s %s",cmdfile->fnames[i],cmdfile->fn_masks[i]);
      if (stat!=2)
	{
	  fprintf(stderr,"Read Error in cmdfile, stat=%d\n\a\n",stat);
	  exit(-1);
	}
    }

  printf("\nInput parameters:\n");
  printf("  cmdfile: %s\n",argv[ct]);
  printf("  number of training images: %d\n",cmdfile->nb_images);
  printf("  regularization factor: %g\n",*reg);
  printf("\n");


}
      

void train_img(struct geom_info *geom,double ***X_true,double ****Z_true,double ****prior_params, int img, int ***mask)
{
  int MODE_FLAG;
  int lev;
  char fnT[255],fnP[255];

  if (img==0)
    MODE_FLAG=NEW;
  else
    MODE_FLAG=APPEND;
  for (lev=PYR_LEVELS-1;lev>0;lev--)
    {
      sprintf(fnT,"tree_dat/train_dat_l%d.asc",lev-1);
      sprintf(fnP,"tree_dat/prune_dat_l%d.asc",lev-1);
      wav_write_train_files(X_true,Z_true, lev, geom, fnT, fnP, MODE_FLAG,mask);
    }
}
 
     
 



