#include "../include/utils.h"
 


/* True corrected mean */
double calc_corrected_mean(double **x,int **mask,int N, int M, int mask_FLAG)
{
  struct entry *array;
  int i,j,len,ct1,ct2,ct;
  double dlen,dct,sum,cr_mean;

  array=(struct entry*)mget_spc(N*M, sizeof(struct entry));
  
  ct=0;

  if (mask_FLAG)
    {
      for (i=0;i<N;i++)
	for (j=0;j<M;j++)
	  if (mask[i][j])
	    {	
	      array[ct].index=ct;
	      array[ct].value=x[i][j];
	      ct++;
	    }
    }
  else
    {
      /*printf("*******************************************************************************NOT USING MASK\n\n");*/
      for (i=0;i<N;i++)
	for (j=0;j<M;j++)
	  {	
	    array[ct].index=ct;
	    array[ct].value=x[i][j];
	    ct++;
	  }
    }




  len=ct;


  quicksort(array,len);


  dlen=(double)len;
  ct1=(int)(0.05*dlen+0.5);
  ct2=(int)(0.95*dlen+0.5);



  
  if (ct1>len-1)
    {
      fprintf(stderr,"Internal error in calc_corrected mean, ct1=%d, len-1=%d\n\a\n",ct1,len-1);
      exit(-1);
    }
  
   if (ct2>len-1)
    {
      fprintf(stderr,"Warning: calc_corrected mean, ct2=%d, len-1=%d, setting ct2=len-1\n\a\n",ct2,len-1);
      ct2=len-1;
    }


   dct=0.0;
   sum=0.0;
   for (i=ct1;i<=ct2;i++)
     {
       sum += array[i].value;
       dct++;
     }

   cr_mean = sum/dct;

   free((void*)array);

   /*printf("cr_mean=%f  ct1=%d  ct2=%d len-1=%d\n",cr_mean,ct1,ct2,len-1);*/

   return(cr_mean);

}






int** read_mask(struct geom_info *geom,char *fname)
{
  MATFile *fin;  
  double **img;
  int **mask;
  int N,M,i,j;

  fin = matOpen(fname, "r");
  if (fin == NULL)
    {
      fprintf(stderr,"Cannot read MatLab file %s\n",fname);
      exit(-1);
    }

  img = read_mat_double_image(&N,&M,"mask",fin);
  if (img == NULL)
    {
      fprintf(stderr,"Cannot read mask matrix \n");
      exit(-1);
    }
  matClose(fin);
  if ((N!=geom->n_y)||(M!=geom->n_x))
    {
      fprintf(stderr,"Size of mask doesn't match size of image\n");
      exit(-1);
    }

  mask=(int**)multialloc(sizeof(int),2,N,M);
  for (i=0;i<N;i++)
    for (j=0;j<M;j++)
      mask[i][j]=(int)(img[i][j]+0.5);
  
  /*printf("**************N=%d M=%d\n",N,M);*/

  free_img((void**)img);
  

  return(mask);
}




void decimate_masks(int ***mask,struct geom_info *geom)
{
  int N,M,i,j,k;

  N=geom->n_y;
  M=geom->n_x;

  for (k=1;k<PYR_LEVELS;k++)
    {
      N/=2;
      M/=2;
      mask[k]=(int**)multialloc(sizeof(int),2,N,M);
      for (i=0;i<N;i++)
	for (j=0;j<M;j++)
	  if (mask[k-1][2*i][2*j]+mask[k-1][2*i][2*j+1]+mask[k-1][2*i+1][2*j]+mask[k-1][2*i+1][2*j+1])
	    mask[k][i][j]=1;
	  else
	    mask[k][i][j]=0;
    }

}
