#include "../include/wav_icm.h"


double wav_ICMstep(struct wavICMdata *wICMdata,struct all_data *data,double *S,struct column *sp_col,int *FLAG)
{	  
  double alpha_new;
  double D1,D2,pri1,pri2;

  *FLAG = wav_compute_derivatives(wICMdata, data, sp_col,S);
  /* This returns negative derivatives actually */
  D1=wICMdata->D1;
  D2=wICMdata->D2;
  pri1=wICMdata->pri1;
  pri2=wICMdata->pri2;

  alpha_new = wICMdata->alpha - (D1+pri2+pri1*wICMdata->alpha)/(D2+pri1);



  return(alpha_new);
}




int wav_compute_derivatives(struct wavICMdata *wICMdata, struct all_data *data, struct column *sp_col,double *S)
{
  int FLAG=0;
  int t,pro;
  double Ypro,Spro;
  double D1,D2,tmp;
  D1=0.0;
  D2=0.0;


  for(t=0;t<sp_col->n_row;t++)
    {
      pro = sp_col->row[t];
      Spro=S[pro];
      Ypro=data->Y[0][pro];


      if ((Spro != 0.0) && (Ypro != 0))
	{
	  
	  D1 += (sp_col->val[t]*(1.0 - Ypro/Spro));
	  tmp = sp_col->val[t]/Spro; 
	  D2 += (Ypro*tmp*tmp);
		      
	}
      else if(Ypro == 0)
	D1 += sp_col->val[t];
      else
	FLAG = 1;
		  
    }
  
 


  wICMdata->D1 = D1;
  wICMdata->D2 = D2;

  return(FLAG);

	
}	







double wav_calc_prior(struct all_data *data,struct geom_info *geom, int recon_lev)
{
  double prior=0.0;
  int i,j,w,m,n;
  static struct wavICMdata wICMdata;
  double ****Z;
  int lev,N,M;

  Z=data->Z;

  for (lev=recon_lev;lev<PYR_LEVELS-1;lev++)
    {
      N=geom[lev].n_y/2;
      M=geom[lev].n_x/2;
      for (i=0;i<N;i++)
	for (j=0;j<M;j++)
	  {
	    /* Load wICMdata with the correct values */
	    /* old wavelet coeff. values and prior means */
	    for (w=0;w<3;w++)
	      {
		wICMdata.V[w]=Z[lev][w][i][j];
		wICMdata.prior_mean[w] = data->allprior_params[lev][w][i][j];
	      }
	    /* prior-covariance matrix */
	    for (m=0;m<3;m++)
	      for (n=0;n<3;n++)
		wICMdata.prior_icov[m][n]= data->allprior_params[lev][3*m+n+3][i][j];
	    
	    /* Now add on pixel prior for this group of 3 wavelet coeffs */
	    prior += wav_pixel_prior(wICMdata.V,&wICMdata);
	    
	  }
    }
  return(prior);
}


 
double wav_pixel_prior(double *V,struct wavICMdata *wICMdata)
{
  double pix_prior=0.0;
  int i,j;


  for (i=0;i<3;i++)
    for (j=0;j<3;j++)
      pix_prior -= ( 0.5*(V[i]-wICMdata->prior_mean[i])*wICMdata->prior_icov[i][j]*(V[j]-wICMdata->prior_mean[j]));

 
  return(pix_prior);

}


void updateS(double *S, struct column *sp_col, double diff)
{
  static int t,pro;
  
  for(t=0;t<sp_col->n_row;t++)
    {
      pro = sp_col->row[t];
      S[sp_col->row[t]] += sp_col->val[t]*diff;

      if (!S[pro]>0.0)
	{
	  fprintf(stderr,"Error in ML recon: S[pro]<0.0\n\a\n");
	  exit(-1);
	}

    }
}



/* based on value of iter, uses different pixel order and returns locations i,j*/
/* for pixel number s*/
void imgcoord(int *out_i,int *out_j,int s,int iter,int N,int M)
{  
  int i,j,ii,jj,MM,NN,itemp;

  if((iter/4)%2 != 0)
    {
      NN = M;
      MM = N;
    }
  else
    {
      NN = N;
      MM = M;
    }
  
  ii=s/MM;
  jj=s-ii*MM;
  
  
  if((iter-1)%4 == 0)
    {
      i = NN-1-ii;
      j = MM-1-jj;
    }
  else 
    if((iter-2)%4 == 0)
      {
	i = NN-1-ii;
	j = jj;
      }
    else 
      if((iter-3)%4 == 0)
	{
	  i = ii;
	  j = MM-1-jj;
	}
      else
	{
	  i = ii;
	  j = jj;
	}
  
  if((iter/4)%2 != 0)
    {
      itemp = i;
      i = j;
      j = itemp;
    }
  
  *out_j=j;
  *out_i=i;

}
 

void  f_pro(double *S,struct all_data *data,struct geom_info *geom,double **profiles,struct column *sp_col, int lev)
{
  int i,j,t;
  int M,N;


  N = geom->n_y;
  M = geom->n_x;
 
  /* Adding scatter */
  for(i=0;i<data->sino.n_t*data->sino.n_theta;i++)
    S[i] = data->scatter;
  
  for(i=0;i<N;i++)
    for(j=0;j<M;j++)
      { 
	A_comp(i,j,&data->sino,geom,profiles,sp_col,lev);
        for(t=0;t<sp_col->n_row; t++)
	  S[sp_col->row[t]] += sp_col->val[t]*data->Xall[lev][i][j];
      }
}


void circ_boundary_conds(int *ic, int *jc, int i, int j, int N, int M)
{
  static int iic,jjc;

  iic=i;
  jjc=j;

  if (iic<0)
    iic+=N;

  if (jjc<0)
    jjc+=M;

  if (iic>=N)
    iic-=N;

  if (jjc>=M)
    jjc-=M;


  *ic=iic;
  *jc=jjc;

}
 

















