#include "../include/wav_recon.h"




double wav_calc_llikelihood(struct all_data *data,struct geom_info *geom, double *S, int recon_lev)
{
  double l_like = 0.0;
  int i;
  double prior;
 
  for(i=0;i<data->sino.n_t*data->sino.n_theta;i++)
    if ( S[i] > 0.0 )
      l_like += (-S[i] + data->Y[0][i]*log( S[i] ));
    else 
      if (data->Y[0][i]>1.0e-10)
	l_like -= 1.0e6;

  prior=wav_calc_prior(data,geom,recon_lev);
  /*printf("l_like=%g  prior=%g",l_like,prior);*/
  l_like += prior;
 
  return(l_like);
}





void calc_prior_params_ls(struct wavICMdata *wICMdata, double **Z_dir,struct all_data *data, int **Z_coord, int recon_lev, struct cluster_info *cluster_info, struct geom_info *geom)
{
  double er[3],d[3];
  double Rinv[3][3];
  double pri1=0.0;
  double pri2=0.0;
  int lev,i,j,n,m;
  double ****prior_params;
  double **a;
  static double Xdir[5]={0.5,0.25,0.125,0.0625,0.0312};
  int cl,ctr_i,ctr_j,ii,jj;
  double dctr;
  double dmu[3];
  int coarse_wdw_pos,N,M;
  int X_len,offs;


  offs=(int)(WINDOW_SIZE/2);
  X_len=WINDOW_SIZE*WINDOW_SIZE-1;


  prior_params=data->allprior_params;
  
  for (lev=recon_lev;lev<PYR_LEVELS-1;lev++)
    {
      ctr_i=Z_coord[lev][0];
      ctr_j=Z_coord[lev][1];
      i=ctr_i;
      j=ctr_j;

      for (n=0;n<3;n++)
	dmu[n]=0.0;

      /* center coordinate of prediction */
      cl=cluster_info->cluster_map[lev][i][j];
      dctr = Xdir[lev-recon_lev];

      if (cl!=-1)
	{
	  /*printf("********** cl=%d\n",cl);*/
	  a=cluster_info->clust[lev][cl].a;
	  
	  /* when pixel changes, it changes center DC-coeff
	     of prediction window for its mean at each resolution 
	     calculate direction of change of predicted prior mean
	     mu=matrix_a' * (vec_xcoarse_outer_window - I*ctr_val_xcoarse_window)
	     where I is identity vector (1x8). 
	     Since vec_xcoarse_outer_window = const when pixel changes,
	     -> dmu = - matrix_a' * (I * dctr)
	  */
	  for (n=0;n<3;n++)
	    for (m=0;m<X_len;m++)
	      dmu[n] -= (dctr*a[m][n]);
	  
	}
      
      for (n=0;n<3;n++)
	{
	  er[n]=data->Z[lev][n][i][j] - prior_params[lev][n][i][j];
	  d[n]=Z_dir[lev][n] - dmu[n];
	}
      
      for (n=0;n<3;n++)
	for (m=0;m<3;m++)
	  Rinv[n][m]=prior_params[lev][3+3*m+n][i][j];
      
      for (n=0;n<3;n++)
	for (m=0;m<3;m++)
	  {
	    pri1 += (d[n]*Rinv[n][m]*d[m]);
	    pri2 += (d[n]*Rinv[n][m]*er[m]);
	  }



      if (cl!=-1)
	{
	  /* add terms for X_len non-center coordinates */
	  N=geom[lev].n_x/2;
	  M=geom[lev].n_y/2;
	  
	  /* i,j below are center coordinates of affected neighboring
	     coarse scale window used for mean prediction for 
	     neighboring wavelet coeff Z[lev][w][i][j] .
	     coarse_wdw_pos is the predictor coordinate of the 
	     wavelet coefficient that we are changing. This coordinate
	     is for the predictor of the NEIGHBORING wavelet coeff,
	     i.e. relative to the center of the prediction window for 
	     that coeff.
	  */
	  coarse_wdw_pos=X_len;
	  for (ii=ctr_i-offs;ii<=ctr_i+offs;ii++)
	    for (jj=ctr_j-offs;jj<=ctr_j+offs;jj++)
	      if ((ii!=ctr_i) || (jj!=ctr_j))
		{
		  coarse_wdw_pos--;
		  i=ii;
		  j=jj;
		  /* circular boundary conds */
		  if (i<0)
		    i+=N;
		  else if (i>N-1)
		    i-=N;
		  if (j<0)
		    j+=M;
		  else if (j>M-1)
		    j-=M;
		  
		  cl=cluster_info->cluster_map[lev][i][j];
		  a=cluster_info->clust[lev][cl].a;
		  for (n=0;n<3;n++)
		    dmu[n] = dctr*a[coarse_wdw_pos][n];
		  
		  for (n=0;n<3;n++)
		    {
		      er[n]=data->Z[lev][n][i][j] - prior_params[lev][n][i][j];
		      d[n]= -dmu[n];
		    }
		  
		  for (n=0;n<3;n++)
		    for (m=0;m<3;m++)
		      Rinv[n][m]=prior_params[lev][3+3*m+n][i][j];
	    
		  for (n=0;n<3;n++)
		    for (m=0;m<3;m++)
		      {
			pri1 += (d[n]*Rinv[n][m]*d[m]);
			pri2 += (d[n]*Rinv[n][m]*er[m]);
		      }
		  
		}
	}
      
    }

  wICMdata->pri1=pri1;
  wICMdata->pri2=pri2;
}




void update_prior_means(struct all_data *data,double diff, int recon_lev, struct cluster_info *cluster_info, int **Z_coord, struct geom_info *geom)
{
  double ****prior_params;
  double **a;
  static double Xdir[5]={0.5,0.25,0.125,0.0625,0.0312};
  double diff_ctr;
  double diff_mu[3];
  int coarse_wdw_pos,N,M;
  int cl,ctr_i,ctr_j;
  int lev,i,j,n,m,ii,jj;
  int X_len,offs;


  offs=(int)(WINDOW_SIZE/2);
  X_len=WINDOW_SIZE*WINDOW_SIZE-1;

  prior_params=data->allprior_params;
  
  for (lev=recon_lev;lev<PYR_LEVELS-1;lev++)
    {
      ctr_i=Z_coord[lev][0];
      ctr_j=Z_coord[lev][1];
      i=ctr_i;
      j=ctr_j;

      /* center coordinate of prediction */
      diff_ctr = diff*Xdir[lev-recon_lev];
      cl=cluster_info->cluster_map[lev][i][j];
      if (cl!=-1)
	{

	  a=cluster_info->clust[lev][cl].a;
	  
	  /* when pixel changes, it changes center DC-coeff
	     of prediction window for its mean at each resolution 
	     calculate direction of change of predicted prior mean
	     mu=matrix_a' * (vec_xcoarse_outer_window - I*ctr_val_xcoarse_window)
	     where I is identity vector (1x8). 
	     Since vec_xcoarse_outer_window = const when pixel changes,
	     -> diff_mu = - matrix_a' * (I * diff_ctr)
	  */
	  for (n=0;n<3;n++)
	    {
	      diff_mu[n]=0.0;
	      for (m=0;m<X_len;m++)
		diff_mu[n] -= (diff_ctr*a[m][n]);
	    }
	  
	  for (n=0;n<3;n++)
	    prior_params[lev][n][i][j] += diff_mu[n];
	  
	  
	  /* add terms for X_len non-center coordinates */
	  N=geom[lev].n_x/2;
	  M=geom[lev].n_y/2;
	  
	  /* i,j below are center coordinates of affected neighboring
	     coarse scale window used for mean prediction for 
	     neighboring wavelet coeff Z[lev][w][i][j] .
	     coarse_wdw_pos is the predictor coordinate of the 
	     wavelet coefficient that we are changing. This coordinate
	     is for the predictor of the NEIGHBORING wavelet coeff,
	     i.e. relative to the center of the prediction window for 
	     that coeff.
	  */
	  coarse_wdw_pos=X_len;
	  for (ii=ctr_i-offs;ii<=ctr_i+offs;ii++)
	    for (jj=ctr_j-offs;jj<=ctr_j+offs;jj++)
	      if ((ii!=ctr_i) || (jj!=ctr_j))
		{
		  coarse_wdw_pos--;
		  i=ii;
		  j=jj;
		  /* circular boundary conds */
		  if (i<0)
		    i+=N;
		  else if (i>N-1)
		    i-=N;
		  if (j<0)
		    j+=M;
		  else if (j>M-1)
		    j-=M;
		  
		  cl=cluster_info->cluster_map[lev][i][j];
		  a=cluster_info->clust[lev][cl].a;
		  for (n=0;n<3;n++)
		    diff_mu[n] = diff_ctr*a[coarse_wdw_pos][n];
		  
		  for (n=0;n<3;n++)
		    prior_params[lev][n][i][j] += diff_mu[n];
		  
		}
	}
      /* why not update Xcoarse at the same time...*/
      data->Xall[lev+1][ctr_i][ctr_j] += diff_ctr;
    }
}



