#include "../include/wav_transform.h"


void wavelet_decomposition(double ****Z,double ***dc,double **Xin,struct geom_info *geom, int recon_lev)
{
  int i,j,k,N,M;
  double pix[4];


  /* updateXZ relies on the fact that DC_dir contains a, so just 
     copy finest scale 
   */
  N=geom[recon_lev].n_y;
  M=geom[recon_lev].n_x;
  for (i=0;i<N;i++)
    for (j=0;j<M;j++)
      dc[recon_lev][i][j]=Xin[i][j];

  /* calc dc for all levels */
  for (k=recon_lev+1;k<PYR_LEVELS;k++)
    {
      N=geom[k].n_y;
      M=geom[k].n_x;
      for (i=0;i<N;i++)
	for (j=0;j<M;j++)
	  dc[k][i][j] = WAV_DC_RECON_MULT*(dc[k-1][2*i][2*j]+dc[k-1][2*i+1][2*j]+dc[k-1][2*i][2*j+1]+dc[k-1][2*i+1][2*j+1]);
    }
  


  for (k=recon_lev;k<PYR_LEVELS-1;k++)
    {
      N=geom[k].n_y;
      M=geom[k].n_x;
      for (i=0;i<N/2;i++)
	for (j=0;j<M/2;j++)
	  {
	    pix[0]=dc[k][2*i][2*j];
	    pix[1]=dc[k][2*i][2*j+1];
	    pix[2]=dc[k][2*i+1][2*j];
	    pix[3]=dc[k][2*i+1][2*j+1];
	    
	    Z[k][0][i][j] = 0.5*(pix[0]+pix[1]-pix[2]-pix[3]);
	    Z[k][1][i][j] = 0.5*(pix[0]-pix[1]+pix[2]-pix[3]);
	    Z[k][2][i][j] = 0.5*(pix[0]-pix[1]-pix[2]+pix[3]);
	  }
    }
}




/* dc_mr must contain coarsest scale dc, then dc_mr and Z are used
   for wavelet recon. to obtain fine scale X. dc_mr will in
   addition conatin dc's at all resolutions incl. finest scale (redundant)
*/
void wavelet_transpose_decomposition(double **X,double ****Z,double ***dc_mr,struct geom_info *geom, int recon_lev)
{
  int i,j,lev,N,M;
  double dc, V[3];

  
  if (fabs(WAV_DC_DECOMP_MULT-0.5)>1e-20)
    {
      fprintf(stderr,"wavelet_transpose_decomposition only implemented for orthonormal wavelet transform\n\a\n");
      exit(-1);
    }
 
 
  /* wavelet reconst. of direction dir to dir_img at finest resolution */
  /* 2. all other levels */
  for (lev=PYR_LEVELS-2;lev>=recon_lev;lev--)
    {
      N=geom[lev].n_y/2;
      M=geom[lev].n_x/2;
      for (i=0;i<N;i++)
	for (j=0;j<M;j++)
	  {
	    V[0]=Z[lev][0][i][j];
	    V[1]=Z[lev][1][i][j];
	    V[2]=Z[lev][2][i][j];
	   
	    dc = WAV_DC_RECON_MULT*dc_mr[lev+1][i][j];

	    dc_mr[lev][2*i][2*j]     = dc + 0.5*V[0] + 0.5*V[1] + 0.5*V[2];
	    dc_mr[lev][2*i][2*j+1]   = dc + 0.5*V[0] - 0.5*V[1] - 0.5*V[2];
	    dc_mr[lev][2*i+1][2*j]   = dc - 0.5*V[0] + 0.5*V[1] - 0.5*V[2];
	    dc_mr[lev][2*i+1][2*j+1] = dc - 0.5*V[0] - 0.5*V[1] + 0.5*V[2];
	  }
    }


  /* copy finest level */
  for (i=0;i<geom[recon_lev].n_y;i++)
    for (j=0;j<geom[recon_lev].n_x;j++)
      X[i][j]=dc_mr[recon_lev][i][j];


  

}

void wavelet_reconstruction(double **X,double ****Z,double ***dc_mr,struct geom_info *geom)
{
  wavelet_transpose_decomposition(X,Z,dc_mr,geom, 0);
}





void wavelet_reconstruction_trdc(double ***X,double ****Z,double ***tr_dc_mr,struct geom_info *geom)
{
  int i,j,lev,N,M;
  double dc, V[3];

    /* wavelet reconst. of direction dir to dir_img at finest resolution */
  /* 2. all other levels */
  for (lev=PYR_LEVELS-2;lev>=0;lev--)
    {
      N=geom[lev].n_y/2;
      M=geom[lev].n_x/2;
      for (i=0;i<N;i++)
	for (j=0;j<M;j++)
	  {
	    V[0]=Z[lev][0][i][j];
	    V[1]=Z[lev][1][i][j];
	    V[2]=Z[lev][2][i][j];
	   
	    dc = WAV_DC_RECON_MULT*tr_dc_mr[lev+1][i][j];

	    X[lev][2*i][2*j]     = dc + 0.5*V[0] + 0.5*V[1] + 0.5*V[2];
	    X[lev][2*i][2*j+1]   = dc + 0.5*V[0] - 0.5*V[1] - 0.5*V[2];
	    X[lev][2*i+1][2*j]   = dc - 0.5*V[0] + 0.5*V[1] - 0.5*V[2];
	    X[lev][2*i+1][2*j+1] = dc - 0.5*V[0] - 0.5*V[1] + 0.5*V[2];
	  }
    }

}

void wav_copy_DC(double ***X,int level_fine,struct geom_info *geom)
{
  int i,j,N,M;
  double dc;

  N=geom[level_fine].n_y/2;
  M=geom[level_fine].n_x/2;
  
  for (i=0;i<N;i++)
    for (j=0;j<M;j++)
      {
	dc = WAV_DC_RECON_MULT*X[level_fine+1][i][j];
	
	X[level_fine][2*i][2*j]     = dc; 
	X[level_fine][2*i][2*j+1]   = dc;
	X[level_fine][2*i+1][2*j]   = dc;
	X[level_fine][2*i+1][2*j+1] = dc;
      }

}




void calc_Z_dir(double **Z_dir,int **Z_coord,int i,int j,int recon_lev)
{
  int lev;
  static int first=1;
  static int mask[PYR_LEVELS-1];
  static double Wt[3][4]={{0.5,0.5,-0.5,-0.5},{0.5,-0.5,0.5,-0.5},{0.5,-0.5,-0.5,0.5}};
  double dc_fact;
  int bit_i,bit_j,wcol,w;




  if (first)
    {
      mask[0]=1;
      for (lev=1;lev<PYR_LEVELS-1;lev++)
	mask[lev] = 2*mask[lev-1];
      first=0;
    }


  /* A single pixel X_ij affects one triplet of wavelet coeffs at
     each resolution (Haar basis). Store locations of these
     triplets at each resolution */
  Z_coord[recon_lev][0] = i/2;
  Z_coord[recon_lev][1] = j/2;
  for (lev=recon_lev+1;lev<PYR_LEVELS-1;lev++)
    {
      Z_coord[lev][0] = Z_coord[lev-1][0] / 2;
      Z_coord[lev][1] = Z_coord[lev-1][1] / 2;
    }

  /* Now find weights with which x_ij affects each wavelet coeff */
  /* in order to find out which one of the four basis pixels x_j
     affects at each resolution, we convert i,j to binary representation. */
  

  if (recon_lev>0)
    for (lev=0;lev<recon_lev;lev++)
      for (w=0;w<3;w++)
	Z_dir[lev][w] = 0.0;
  
  dc_fact=1.0;
  for (lev=recon_lev;lev<PYR_LEVELS-1;lev++)
    {  
      bit_i=(i & mask[lev-recon_lev]) ? 1 : 0;
      bit_j=(j & mask[lev-recon_lev]) ? 1 : 0;

      wcol=2*bit_i+bit_j;
      /* i.e. wcol=0: upper left pix, wcol=1: upper right pix, 
	 wcol=2: lower left, wcol=3: lower right */

      for (w=0;w<3;w++)
	Z_dir[lev][w] = dc_fact*Wt[w][wcol];
      
      dc_fact *= 0.5;
    }


}
